# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""QuerySet classes for DjangoCA models."""

import abc
import typing
from typing import TYPE_CHECKING, Any, Generic, List, TypeVar

from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.db.models import Q
from django.utils import timezone

from django_ca import ca_settings
from django_ca.acme.constants import Status
from django_ca.typehints import X509CertMixinTypeVar
from django_ca.utils import sanitize_serial

if not TYPE_CHECKING:
    # Inverting TYPE_CHECKING check here to make pylint==2.9.3 happy:
    #   https://github.com/PyCQA/pylint/issues/4697
    AcmeAccountQuerySetBase = (
        AcmeAuthorizationQuerySetBase
    ) = (
        AcmeCertificateQuerySetBase
    ) = (
        AcmeChallengeQuerySetBase
    ) = AcmeOrderQuerySetBase = CertificateQuerySetBase = CertificateAuthorityQuerySetBase = models.QuerySet

    QuerySetTypeVar = TypeVar("QuerySetTypeVar", bound=models.QuerySet)
else:  # pragma: no cover  # only used for type checking
    from django_ca.models import (
        AcmeAccount,
        AcmeAuthorization,
        AcmeCertificate,
        AcmeChallenge,
        AcmeOrder,
        Certificate,
        CertificateAuthority,
        X509CertMixin,
    )

    AcmeAccountQuerySetBase = models.QuerySet[AcmeAccount]
    AcmeAuthorizationQuerySetBase = models.QuerySet[AcmeAuthorization]
    AcmeCertificateQuerySetBase = models.QuerySet[AcmeCertificate]
    AcmeChallengeQuerySetBase = models.QuerySet[AcmeChallenge]
    AcmeOrderQuerySetBase = models.QuerySet[AcmeOrder]
    CertificateAuthorityQuerySetBase = models.QuerySet[CertificateAuthority]
    CertificateQuerySetBase = models.QuerySet[Certificate]

    QuerySetTypeVar = TypeVar("QuerySetTypeVar", bound=models.QuerySet[X509CertMixin])


class QuerySetProtocol(
    typing.Protocol[X509CertMixinTypeVar]
):  # pragma: nocover; pylint: disable=missing-function-docstring
    """Protocol used for a generic-self in mixins.

    Note that I couldn't get this to work in functions that should return the same type as well. So::

        def filter(self: QuerySetProtocol) -> QuerySetProtocol:
            ...

    ... doesn't work, unfortunately.

    .. seealso:: https://mypy.readthedocs.io/en/latest/more_types.html#mixin-classes
    """

    model: X509CertMixinTypeVar

    def get(self, *args: Any, **kwargs: Any) -> X509CertMixinTypeVar:
        ...


class DjangoCAMixin(Generic[X509CertMixinTypeVar], metaclass=abc.ABCMeta):
    """Mixin with common methods for CertificateAuthority and Certificate models."""

    def get_by_serial_or_cn(
        self: QuerySetProtocol[X509CertMixinTypeVar], identifier: str
    ) -> X509CertMixinTypeVar:
        """Get a model by serial *or* by common name.

        This method is meant to get a CA from a user input value. If `identifier` is a serial, colons (``:``)
        and leading zeros are ignored. If no exact match is found it will search for CAs starting with that
        value. For example, if a CA has the serial ``ABCDE``, it will be found with "ABCDE", "A:BC:DE",
        "0A:BC:DE" or just "0AB" as `identifier`.
        """
        identifier = identifier.strip()
        exact_query = startswith_query = Q(cn=identifier)

        try:
            serial = sanitize_serial(identifier)
            exact_query |= Q(serial=serial)
            startswith_query |= Q(serial__startswith=serial)
        except ValueError:
            pass

        try:
            # Imported CAs might have a shorter serial and there is a chance that it might become impossible
            # to select a CA by serial if its serial matches another CA with a longer serial. So we try to
            # match by exact serial first.
            return self.get(exact_query)
        except self.model.DoesNotExist:
            return self.get(startswith_query)


class CertificateAuthorityQuerySet(DjangoCAMixin["CertificateAuthority"], CertificateAuthorityQuerySetBase):
    """QuerySet for the CertificateAuthority model."""

    def acme(self) -> "CertificateAuthorityQuerySet":
        """Return usable CAs that have support for the ACME protocol enabled."""
        return self.filter(acme_enabled=True)

    def default(self) -> "CertificateAuthority":
        """Return the default CA to use when no CA is selected.

        This function honors the :ref:`CA_DEFAULT_CA <settings-ca-default-ca>`. If no usable CA can be
        returned, raises :py:exc:`~django:django.core.exceptions.ImproperlyConfigured`.

        Raises
        ------

        :py:exc:`~django:django.core.exceptions.ImproperlyConfigured`
            When the CA named by :ref:`CA_DEFAULT_CA <settings-ca-default-ca>` is either not found, disabled
            or not currently valid. Or, if the setting is not set, no CA is currently usable.
        """

        if ca_settings.CA_DEFAULT_CA:
            now = timezone.now()

            try:
                # NOTE: Don't prefilter queryset so that we can provide more specialized error messages below.
                ca = self.get(serial=ca_settings.CA_DEFAULT_CA)
            except self.model.DoesNotExist:
                # pylint: disable=raise-missing-from; not useful here
                raise ImproperlyConfigured(f"CA_DEFAULT_CA: {ca_settings.CA_DEFAULT_CA}: CA not found.")

            if ca.enabled is False:
                raise ImproperlyConfigured(f"CA_DEFAULT_CA: {ca_settings.CA_DEFAULT_CA} is disabled.")
            if ca.expires < now:
                raise ImproperlyConfigured(f"CA_DEFAULT_CA: {ca_settings.CA_DEFAULT_CA} is expired.")
            if ca.valid_from > now:  # OK, how could this ever happen? ;-)
                raise ImproperlyConfigured(f"CA_DEFAULT_CA: {ca_settings.CA_DEFAULT_CA} is not yet valid.")
            return ca

        # NOTE: We add the serial to sorting make *sure* we have deterministic behavior. In many cases, users
        # will just create several CAs that all actually expire on the same day.
        first_ca = self.usable().order_by("-expires", "serial").first()  # usable == enabled and valid
        if first_ca is None:
            raise ImproperlyConfigured("No CA is currently usable.")
        return first_ca

    def disabled(self) -> "CertificateAuthorityQuerySet":
        """Return CAs that are disabled."""
        return self.filter(enabled=False)

    def enabled(self) -> "CertificateAuthorityQuerySet":
        """Return CAs that are enabled."""
        return self.filter(enabled=True)

    def valid(self) -> "CertificateAuthorityQuerySet":
        """Return CAs that are currently valid."""
        now = timezone.now()
        return self.filter(expires__gt=now, valid_from__lt=now)

    def invalid(self) -> "CertificateAuthorityQuerySet":
        """Return CAs that are either expired or not yet valid."""
        now = timezone.now()
        return self.exclude(expires__gt=now, valid_from__lt=now)

    def revoked(self) -> "CertificateAuthorityQuerySet":
        """Return revoked certificates."""

        return self.filter(revoked=True)

    def usable(self) -> "CertificateAuthorityQuerySet":
        """Return CAs that are enabled and currently valid."""
        return self.enabled().valid()


class CertificateQuerySet(DjangoCAMixin["Certificate"], CertificateQuerySetBase):
    """QuerySet for the Certificate model."""

    def currently_valid(self) -> "CertificateQuerySet":
        """Return certificates currently valid according to their not_before/not_after fields.

        .. WARNING:: This does not exclude revoked certificates.
        """
        now = timezone.now()
        return self.filter(expires__gt=now, valid_from__lt=now)

    def not_yet_valid(self) -> "CertificateQuerySet":
        """Return certificates that are not yet valid."""

        return self.filter(revoked=False, valid_from__gt=timezone.now())

    def valid(self) -> "CertificateQuerySet":
        """Return valid certificates."""

        return self.currently_valid().filter(revoked=False)

    def expired(self) -> "CertificateQuerySet":
        """Returns expired certificates.

        Note that this method does not return revoked certificates that would otherwise be expired.
        """
        return self.filter(revoked=False, expires__lt=timezone.now())

    def revoked(self) -> "CertificateQuerySet":
        """Return revoked certificates."""

        return self.filter(revoked=True)


class AcmeAccountQuerySet(AcmeAccountQuerySetBase):
    """QuerySet for :py:class:`~django_ca.models.AcmeAccount`."""

    def viewable(self) -> "AcmeAccountQuerySet":
        """Filter ACME accounts that can be viewed via the ACME API.

        An account is considered viewable if the associated CA is usable. Note that an account is *viewable*
        also if it was revoked by the CA.
        """
        now = timezone.now()
        return self.filter(
            ca__enabled=True, ca__acme_enabled=True, ca__expires__gt=now, ca__valid_from__lt=now
        )


class AcmeOrderQuerySet(AcmeOrderQuerySetBase):
    """QuerySet for :py:class:`~django_ca.models.AcmeOrder`."""

    def account(self, account: "AcmeAccount") -> "AcmeOrderQuerySet":
        """Filter orders belonging to the given account."""
        return self.filter(account=account)

    def viewable(self) -> "AcmeOrderQuerySet":
        """Filter ACME orders that can be viewed via the ACME API.

        An order is considered viewable if the associated CA is usable and the account is not revoked.
        """
        now = timezone.now()
        return self.filter(
            account__ca__enabled=True,
            account__ca__acme_enabled=True,
            account__ca__expires__gt=now,
            account__ca__valid_from__lt=now,
        ).exclude(account__status=Status.REVOKED.value)


class AcmeAuthorizationQuerySet(AcmeAuthorizationQuerySetBase):
    """QuerySet for :py:class:`~django_ca.models.AcmeAuthorization`."""

    def account(self, account: "AcmeAccount") -> "AcmeAuthorizationQuerySet":
        """Filter authorizations belonging to the given account."""
        return self.filter(order__account=account)

    def dns(self) -> "AcmeAuthorizationQuerySet":
        """Get all authorizations of type DNS."""
        return self.filter(type=self.model.TYPE_DNS)

    def names(self) -> List[str]:
        """Get a flat list of names identified by the current queryset."""
        return list(self.values_list("value", flat=True))

    def url(self) -> "AcmeAuthorizationQuerySet":
        """Prepare queryset to get the ACME URL of objects without subsequent database lookups."""
        return self.select_related("order__account__ca")

    def valid(self) -> "AcmeAuthorizationQuerySet":
        """Filter for currently valid authorizations."""
        return self.filter(order__expires__gt=timezone.now(), status=self.model.STATUS_VALID)

    def viewable(self) -> "AcmeAuthorizationQuerySet":
        """Filter ACME authzs that can be viewed via the ACME API.

        An authz is considered viewable if the associated CA is usable and the account is not revoked.
        """
        now = timezone.now()
        return self.filter(
            order__account__ca__enabled=True,
            order__account__ca__acme_enabled=True,
            order__account__ca__expires__gt=now,
            order__account__ca__valid_from__lt=now,
        ).exclude(order__account__status=Status.REVOKED.value)


class AcmeChallengeQuerySet(AcmeChallengeQuerySetBase):
    """QuerySet for :py:class:`~django_ca.models.AcmeChallenge`."""

    def account(self, account: "AcmeAccount") -> "AcmeChallengeQuerySet":
        """Filter challenges belonging to the given account."""
        return self.filter(auth__order__account=account)

    def url(self) -> "AcmeChallengeQuerySet":
        """Prepare queryset to get the ACME URL of objects without subsequent database lookups."""
        return self.select_related("auth__order__account__ca")

    def viewable(self) -> "AcmeChallengeQuerySet":
        """Filter ACME challenges that can be viewed via the ACME API.

        An authz is considered viewable if the associated CA is usable and the account is not revoked.
        """
        now = timezone.now()
        return self.filter(
            auth__order__account__ca__enabled=True,
            auth__order__account__ca__acme_enabled=True,
            auth__order__account__ca__expires__gt=now,
            auth__order__account__ca__valid_from__lt=now,
        ).exclude(auth__order__account__status=Status.REVOKED.value)


class AcmeCertificateQuerySet(AcmeCertificateQuerySetBase):
    """QuerySet for :py:class:`~django_ca.models.AcmeCertificate`."""

    def account(self, account: "AcmeAccount") -> "AcmeCertificateQuerySet":
        """Filter certificates belonging to the given account."""
        return self.filter(order__account=account)

    def url(self) -> "AcmeCertificateQuerySet":
        """Prepare queryset to get the ACME URL of objects without subsequent database lookups."""
        return self.select_related("order__account__ca")

    def viewable(self) -> "AcmeCertificateQuerySet":
        """Filter ACME certificates that can be viewed via the ACME API.

        An authz is considered viewable if the associated CA is usable, the order is ready, the account is not
        revoked and the certificate itself was not revoked.
        """
        now = timezone.now()
        return (
            self.filter(
                order__account__ca__enabled=True,
                order__account__ca__acme_enabled=True,
                order__account__ca__expires__gt=now,
                order__account__ca__valid_from__lt=now,
                order__status=Status.VALID.value,
            )
            .exclude(order__account__status=Status.REVOKED.value)
            .exclude(cert__isnull=True)
            .exclude(cert__revoked=True)
        )



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Management command to sign a new certificate.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

import sys
from datetime import timedelta
from typing import Any, List, Optional

from cryptography import x509
from cryptography.x509.oid import ExtensionOID, NameOID

from django.core.management.base import CommandError, CommandParser
from django.utils import timezone

from django_ca import ca_settings, constants
from django_ca.management.base import BaseSignCertCommand
from django_ca.models import Certificate, CertificateAuthority, Watcher
from django_ca.profiles import profiles
from django_ca.typehints import AllowedHashTypes, ExtensionMapping


class Command(BaseSignCertCommand):  # pylint: disable=missing-class-docstring
    help = f"""Sign a CSR and output signed certificate. The defaults depend on the configured
default profile, currently {ca_settings.CA_DEFAULT_PROFILE}."""

    add_extensions_help = """Values for more complex x509 extensions. This is for advanced usage only, the
profiles already set the correct values for the most common use cases. See
https://django-ca.readthedocs.io/en/latest/extensions.html for more information."""
    subject_help = """The certificate subject of the CSR is not used. The default subject is configured
            with the CA_DEFAULT_SUBJECT setting and may be overwritten by a profile named with
            --profile. The --subject option allows you to name a CommonName (which is not usually
            in the defaults) and override any default values."""

    def add_cn_in_san(self, parser: CommandParser) -> None:
        """Add argument group for the CommonName-in-SubjectAlternativeName options."""
        if ca_settings.CA_PROFILES[ca_settings.CA_DEFAULT_PROFILE]["cn_in_san"]:
            cn_in_san_default = " (default)"
            cn_not_in_san_default = ""
        else:
            cn_in_san_default = ""
            cn_not_in_san_default = " (default)"

        group = parser.add_argument_group(
            "CommonName in subjectAltName",
            """Whether or not to automatically include the CommonName (given in --subject) in the
            list of subjectAltNames (given by --alt).""",
        )
        group = group.add_mutually_exclusive_group()

        group.add_argument(
            "--cn-not-in-san",
            default=None,
            action="store_false",
            dest="cn_in_san",
            help=f"Do not add the CommonName as subjectAlternativeName{cn_not_in_san_default}.",
        )
        group.add_argument(
            "--cn-in-san",
            default=None,
            action="store_true",
            dest="cn_in_san",
            help=f"Add the CommonName as subjectAlternativeName{cn_in_san_default}.",
        )

    def add_arguments(self, parser: CommandParser) -> None:
        general_group = self.add_base_args(parser)
        self.add_cn_in_san(parser)

        general_group.add_argument(
            "--csr",
            dest="csr_path",
            default="-",
            metavar="FILE",
            help="The path to the certificate to sign, if omitted, you will be be prompted.",
        )
        general_group.add_argument(
            "-b", "--bundle", default=False, action="store_true", help="Output the whole certificate bundle."
        )

        self.add_profile(
            parser,
            """Sign certificate based on the given profile. A profile only sets the the
                         default values, options like --key-usage still override the profile.""",
        )

    def handle(  # pylint: disable=too-many-arguments,too-many-locals,too-many-branches,too-many-statements
        self,
        ca: CertificateAuthority,
        subject: Optional[x509.Name],
        expires: Optional[timedelta],
        watch: List[str],
        password: Optional[bytes],
        cn_in_san: bool,
        csr_path: str,
        bundle: bool,
        profile: Optional[str],
        out: Optional[str],
        algorithm: Optional[AllowedHashTypes],
        # Authority Information Access extension
        authority_information_access: x509.AuthorityInformationAccess,
        # Certificate Policies extension
        certificate_policies: Optional[x509.CertificatePolicies],
        certificate_policies_critical: bool,
        # CRL Distribution Points extension
        crl_full_names: Optional[List[x509.GeneralName]],
        crl_distribution_points_critical: bool,
        # Extended Key Usage extension
        extended_key_usage: Optional[x509.ExtendedKeyUsage],
        extended_key_usage_critical: bool,
        # Issuer Alternative Name extension:
        issuer_alternative_name: Optional[x509.IssuerAlternativeName],
        # Key Usage extension
        key_usage: Optional[x509.KeyUsage],
        key_usage_critical: bool,
        # OCSP No Check extension
        ocsp_no_check: bool,
        ocsp_no_check_critical: bool,
        # Subject Alternative Name extension
        subject_alternative_name: Optional[x509.SubjectAlternativeName],
        subject_alternative_name_critical: bool,
        # TLSFeature extension
        tls_feature: Optional[x509.TLSFeature],
        tls_feature_critical: bool,
        **options: Any,
    ) -> None:
        # Validate parameters early so that we can return better feedback to the user.
        if ca.expires < timezone.now():
            raise CommandError("Certificate Authority has expired.")
        if ca.revoked:
            raise CommandError("Certificate Authority is revoked.")

        # Get/validate signature hash algorithm
        algorithm = self.get_hash_algorithm(ca.key_type, algorithm, ca.algorithm)

        profile_obj = profiles[profile]
        self.test_options(ca=ca, expires=expires, password=password, profile=profile_obj, **options)

        # get list of watchers
        watchers = [Watcher.from_addr(addr) for addr in watch]

        # Process any extensions given via the command-line
        extensions: ExtensionMapping = {}

        if authority_information_access is not None:
            self._add_extension(
                extensions,
                authority_information_access,
                constants.EXTENSION_DEFAULT_CRITICAL[ExtensionOID.AUTHORITY_INFORMATION_ACCESS],
            )
        if certificate_policies is not None:
            self._add_extension(extensions, certificate_policies, certificate_policies_critical)
        if crl_full_names is not None:
            distribution_point = x509.DistributionPoint(
                full_name=crl_full_names, relative_name=None, crl_issuer=None, reasons=None
            )
            self._add_extension(
                extensions, x509.CRLDistributionPoints([distribution_point]), crl_distribution_points_critical
            )
        if extended_key_usage is not None:
            self._add_extension(extensions, extended_key_usage, extended_key_usage_critical)
        if issuer_alternative_name is not None:
            self._add_extension(
                extensions,
                issuer_alternative_name,
                constants.EXTENSION_DEFAULT_CRITICAL[ExtensionOID.ISSUER_ALTERNATIVE_NAME],
            )
        if key_usage is not None:
            self._add_extension(extensions, key_usage, key_usage_critical)
        if ocsp_no_check is True:
            self._add_extension(extensions, x509.OCSPNoCheck(), ocsp_no_check_critical)
        if subject_alternative_name is not None:
            self._add_extension(extensions, subject_alternative_name, subject_alternative_name_critical)
        if tls_feature is not None:
            self._add_extension(extensions, tls_feature, tls_feature_critical)

        cname = None
        if subject is not None:
            cname = subject.get_attributes_for_oid(NameOID.COMMON_NAME)
        if not cname and ExtensionOID.SUBJECT_ALTERNATIVE_NAME not in extensions:
            raise CommandError(
                "Must give at least a Common Name in --subject or one or more "
                "--subject-alternative-name/--name arguments."
            )

        # Read the CSR
        if csr_path == "-":
            self.stdout.write("Please paste the CSR:")
            csr_bytes = b""
            while True:
                csr_bytes += sys.stdin.buffer.read(1)
                # COVERAGE NOTE: mock function always returns the full string, so we always break right away
                if csr_bytes.strip().endswith(b"-----END CERTIFICATE REQUEST-----"):  # pragma: no branch
                    break
        else:
            with open(csr_path, "rb") as csr_stream:
                csr_bytes = csr_stream.read()

        if csr_bytes.startswith(b"-----BEGIN CERTIFICATE REQUEST-----"):
            csr = x509.load_pem_x509_csr(csr_bytes)
        else:
            csr = x509.load_der_x509_csr(csr_bytes)

        try:
            cert = Certificate.objects.create_cert(
                ca,
                csr,
                profile=profile_obj,
                cn_in_san=cn_in_san,
                expires=expires,
                extensions=extensions.values(),
                password=password,
                subject=subject,
                algorithm=algorithm,
            )
        except Exception as ex:
            raise CommandError(ex) from ex

        cert.watchers.add(*watchers)

        if bundle is True:
            output = cert.bundle_as_pem
        else:
            output = cert.pub.pem

        if out:
            with open(out, "w", encoding="ascii") as stream:
                stream.write(output)
        else:
            self.stdout.write(output)



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Various type aliases used in throughout django-ca."""

import argparse
import typing
from datetime import datetime, timedelta
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.x509.certificate_transparency import SignedCertificateTimestamp

from django.core.management.base import CommandParser

# Module level imports to enable forward references. See also:
#
#   https://peps.python.org/pep-0484/#forward-references
if typing.TYPE_CHECKING:
    from django_ca import models


class SupportsLessThan(typing.Protocol):
    """Protocol that specifies <, making something sortable."""

    def __lt__(self, __other: Any) -> bool:  # pragma: nocover
        ...


# pylint: disable-next=invalid-name
JSON = Union[Dict[str, "JSON"], List["JSON"], str, int, float, bool, None]

# from cryptography.hazmat.primitives.asymmetric.types import CertificateIssuerPrivateKeyTypes

CRLExtensionTypeTypeVar = typing.TypeVar(
    "CRLExtensionTypeTypeVar", x509.CRLDistributionPoints, x509.FreshestCRL
)

#: Hash algorithms that can be used for signing certificates.
#: NOTE: This is a duplicate of the protected ``cryptography.x509.base._AllowedHashTypes``.
AllowedHashTypes = typing.Union[
    hashes.SHA224,
    hashes.SHA256,
    hashes.SHA384,
    hashes.SHA512,
    hashes.SHA3_224,
    hashes.SHA3_256,
    hashes.SHA3_384,
    hashes.SHA3_512,
]

ExtensionDict = Dict[x509.ObjectIdentifier, x509.Extension[x509.ExtensionType]]

KeyUsages = typing.Literal[
    "cRLSign",
    "dataEncipherment",
    "decipherOnly",
    "digitalSignature",
    "encipherOnly",
    "keyAgreement",
    "keyCertSign",
    "keyEncipherment",
    "nonRepudiation",  # http://marc.info/?t=107176106300005&r=1&w=2
]


ParsableName = Union[str, Iterable[Tuple[str, str]]]

Expires = Optional[Union[int, datetime, timedelta]]
ParsableHash = Optional[Union[str, AllowedHashTypes]]
ParsableKeyType = typing.Literal["RSA", "DSA", "EC", "Ed25519", "Ed448"]
ParsableSubject = Union[
    str,
    # Union for keys is not supported, see: https://github.com/python/mypy/issues/6001
    typing.Mapping[x509.ObjectIdentifier, Union[str, Iterable[str]]],
    typing.Mapping[str, Union[str, Iterable[str]]],
    x509.Name,
    Iterable[Tuple[Union[x509.ObjectIdentifier, str], Union[str, Iterable[str]]]],
]

# GeneralNameList
ParsableGeneralName = Union[x509.GeneralName, str]
ParsableGeneralNameList = Iterable[ParsableGeneralName]

SerializedExtension = typing.TypedDict(
    "SerializedExtension",
    {
        "critical": bool,
        # Value should be a generic typevar, but this is not yet supported in mypy:
        #   https://github.com/python/mypy/issues/3863
        "value": Any,
    },
)
SerializedObjectIdentifier = typing.TypedDict(
    "SerializedObjectIdentifier",
    {
        "oid": str,
        "value": str,
    },
)
SerializedName = List[SerializedObjectIdentifier]
SerializedProfile = typing.TypedDict(
    "SerializedProfile",
    {
        "cn_in_san": bool,
        "description": str,
        "subject": Optional[SerializedName],
        "extensions": Dict[str, Any],
    },
)


# Looser variants of the above for incoming arguments
ParsableNoticeReference = typing.TypedDict(
    "ParsableNoticeReference", {"organization": str, "notice_numbers": Iterable[int]}, total=False
)
ParsableUserNotice = typing.TypedDict(
    "ParsableUserNotice",
    {"notice_reference": Union[x509.NoticeReference, ParsableNoticeReference], "explicit_text": str},
    total=False,
)

# Parsable arguments
ParsableDistributionPoint = typing.TypedDict(
    "ParsableDistributionPoint",
    {
        "full_name": Optional[ParsableGeneralNameList],
        "relative_name": Union[str, x509.RelativeDistinguishedName],
        "crl_issuer": ParsableGeneralNameList,
        "reasons": Iterable[Union[str, x509.ReasonFlags]],
    },
    total=False,
)
ParsablePolicyQualifier = Union[str, x509.UserNotice, ParsableUserNotice]
ParsablePolicyIdentifier = Union[str, x509.ObjectIdentifier]
ParsablePolicyInformation = typing.TypedDict(
    "ParsablePolicyInformation",
    {
        "policy_identifier": ParsablePolicyIdentifier,
        "policy_qualifiers": Optional[typing.Sequence[ParsablePolicyQualifier]],
    },
    total=False,
)

PolicyQualifier = Union[str, x509.UserNotice]

ExtensionTypeTypeVar = typing.TypeVar("ExtensionTypeTypeVar", bound=x509.ExtensionType)
# """A type variable for a :py:class:`~cg:cryptography.x509.ExtensionType` instance."""

ParsableExtension = typing.TypedDict(
    "ParsableExtension",
    {
        "critical": bool,
        # Value should be a generic typevar, but this is not yet supported in mypy:
        #   https://github.com/python/mypy/issues/3863
        "value": Any,
    },
    total=False,
)

if typing.TYPE_CHECKING:
    ExtensionTypeVar = x509.Extension[ExtensionTypeTypeVar]
    ExtensionType = x509.Extension[x509.ExtensionType]
    SubjectKeyIdentifierType = x509.Extension[x509.SubjectKeyIdentifier]
    UnrecognizedExtensionType = x509.Extension[x509.UnrecognizedExtension]
    TLSFeatureExtensionType = x509.Extension[x509.TLSFeature]
    PrecertificateSignedCertificateTimestampsType = x509.Extension[
        x509.PrecertificateSignedCertificateTimestamps
    ]
else:
    ExtensionType = ExtensionTypeVar = x509.Extension
    SubjectKeyIdentifierType = (
        TLSFeatureExtensionType
    ) = UnrecognizedExtensionType = PrecertificateSignedCertificateTimestampsType = x509.ExtensionType


BasicConstraintsBase = typing.TypedDict("BasicConstraintsBase", {"ca": bool})
ParsableAuthorityKeyIdentifierDict = typing.TypedDict(
    "ParsableAuthorityKeyIdentifierDict",
    {
        "key_identifier": Optional[bytes],
        "authority_cert_issuer": Iterable[str],
        "authority_cert_serial_number": Optional[int],
    },
    total=False,
)


SerializedNullExtension = typing.TypedDict("SerializedNullExtension", {"critical": bool})

################
# Type aliases #
################
ExtensionMapping = Dict[x509.ObjectIdentifier, x509.Extension[x509.ExtensionType]]

# Type aliases for protected subclass returned by add_argument_group().
ArgumentGroup = argparse._ArgumentGroup  # pylint: disable=protected-access
# An CommandParser (subclass of argparse.ArgumentParser) or an argument group added by add_argument_group().
ActionsContainer = Union[CommandParser, ArgumentGroup]

############
# TypeVars #
############
# pylint: disable-next=invalid-name  # Should match class, but pylint is more sensitive here
X509CertMixinTypeVar = typing.TypeVar("X509CertMixinTypeVar", bound="models.X509CertMixin")


#####################
# Serialized values #
#####################
# Collect JSON-serializable versions of cryptography values. Typehints in this section start with
# "Serialized...".
SerializedAuthorityInformationAccess = typing.TypedDict(
    "SerializedAuthorityInformationAccess",
    {
        "issuers": List[str],
        "ocsp": List[str],
    },
    total=False,
)
SerializedAuthorityKeyIdentifier = typing.TypedDict(
    "SerializedAuthorityKeyIdentifier",
    {
        "key_identifier": str,
        "authority_cert_issuer": List[str],
        "authority_cert_serial_number": int,
    },
    total=False,
)


# pylint: disable-next=inherit-non-class; False positive
class SerializedBasicConstraints(BasicConstraintsBase, total=False):
    """Serialized representation of a BasicConstraints extension.

    A value of this type is a dictionary with a ``"ca"`` key with a boolean value. If ``True``, it also
    has a ``"path_length"`` value that is either ``None`` or an int.
    """

    path_length: Optional[int]


SerializedDistributionPoint = typing.TypedDict(
    "SerializedDistributionPoint",
    {
        "full_name": List[str],
        "relative_name": str,
        "crl_issuer": List[str],
        "reasons": List[str],
    },
    total=False,
)
SerializedDistributionPoints = typing.TypedDict(
    "SerializedDistributionPoints",
    {
        "critical": bool,
        "value": List[SerializedDistributionPoint],
    },
)

SerializedNameConstraints = typing.TypedDict(
    "SerializedNameConstraints",
    {
        "permitted": List[str],
        "excluded": List[str],
    },
    total=False,
)
SerializedPolicyConstraints = typing.TypedDict(
    "SerializedPolicyConstraints",
    {
        "inhibit_policy_mapping": int,
        "require_explicit_policy": int,
    },
    total=False,
)

# PolicyInformation serialization
SerializedNoticeReference = typing.TypedDict(
    "SerializedNoticeReference", {"organization": str, "notice_numbers": List[int]}, total=False
)
SerializedUserNotice = typing.TypedDict(
    "SerializedUserNotice", {"explicit_text": str, "notice_reference": SerializedNoticeReference}, total=False
)
SerializedPolicyQualifier = Union[str, SerializedUserNotice]
SerializedPolicyQualifiers = List[SerializedPolicyQualifier]
SerializedPolicyInformation = typing.TypedDict(
    "SerializedPolicyInformation",
    {"policy_identifier": str, "policy_qualifiers": Optional[SerializedPolicyQualifiers]},
)

SerializedSignedCertificateTimestamp = typing.TypedDict(
    "SerializedSignedCertificateTimestamp",
    {
        "log_id": str,
        "timestamp": str,
        "type": str,
        "version": str,
    },
)
"""A dictionary with four keys: log_id, timestamp, type, version, values are all str."""

###################
# Parsable values #
###################
# Collect typehints for values that can be parsed back into cryptography values. Typehints in this section
# start with "Parsable...".

ParsableAuthorityKeyIdentifier = Union[str, bytes, ParsableAuthorityKeyIdentifierDict]
ParsableAuthorityInformationAccess = typing.TypedDict(
    "ParsableAuthorityInformationAccess",
    {
        "ocsp": Optional[ParsableGeneralNameList],
        "issuers": Optional[ParsableGeneralNameList],
    },
    total=False,
)


# pylint: disable-next=inherit-non-class; False positive
class ParsableBasicConstraints(BasicConstraintsBase, total=False):
    """Serialized representation of a BasicConstraints extension.

    A value of this type is a dictionary with a ``"ca"`` key with a boolean value. If ``True``, it also
    has a ``"path_length"`` value that is either ``None`` or an int.
    """

    path_length: Union[int, str]


ParsableNameConstraints = typing.TypedDict(
    "ParsableNameConstraints",
    {
        "permitted": ParsableGeneralNameList,
        "excluded": ParsableGeneralNameList,
    },
    total=False,
)

ParsablePolicyConstraints = typing.TypedDict(
    "ParsablePolicyConstraints",
    {
        "require_explicit_policy": int,
        "inhibit_policy_mapping": int,
    },
    total=False,
)
ParsableSignedCertificateTimestamp = Union[SerializedSignedCertificateTimestamp, SignedCertificateTimestamp]
ParsableSubjectKeyIdentifier = Union[str, bytes, x509.SubjectKeyIdentifier]


#####################
# Type alternatives #
#####################
# Collect Union[] typehints that occur multiple times, e.g. multiple x509.ExtensionType classes that behave
# the same way. Typehints in this section are named "...Type".

AlternativeNameExtensionType = Union[x509.SubjectAlternativeName, x509.IssuerAlternativeName]
CRLExtensionType = Union[x509.FreshestCRL, x509.CRLDistributionPoints]
InformationAccessExtensionType = Union[x509.AuthorityInformationAccess, x509.SubjectInformationAccess]
SignedCertificateTimestampType = Union[
    x509.PrecertificateSignedCertificateTimestamps, x509.SignedCertificateTimestamps
]



# Generated by Django 4.2.3 on 2023-07-16 11:01

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_ca", "0031_certificateauthority_sign_certificate_policies"),
    ]

    operations = [
        migrations.AddField(
            model_name="certificateauthority",
            name="ocsp_responder_key_validity",
            field=models.PositiveSmallIntegerField(
                default=3,
                help_text="How long <strong>(in days)</strong> OCSP responder keys may be valid.",
                validators=[django.core.validators.MinValueValidator(1)],
                verbose_name="OCSP responder key validity",
            ),
        ),
        migrations.AddField(
            model_name="certificateauthority",
            name="ocsp_response_validity",
            field=models.PositiveIntegerField(
                default=86400,
                help_text="How long <strong>(in seconds)</strong> OCSP responses may be considered valid by the client.",
                validators=[django.core.validators.MinValueValidator(600)],
                verbose_name="OCSP response validity",
            ),
        ),
    ]



# Generated by Django 2.2.1 on 2019-05-18 10:46

from django.db import migrations


def remove_empty(apps, schema_editor):
    Certificate = apps.get_model('django_ca', 'Certificate')
    Certificate.objects.filter(revoked_reason='').update(revoked_reason='unspecified')
    CertificateAuthority = apps.get_model('django_ca', 'CertificateAuthority')
    CertificateAuthority.objects.filter(revoked_reason='').update(revoked_reason='unspecified')


def noop():
    pass


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0013_certificateauthority_crl_number'),
    ]

    operations = [
        migrations.RunPython(remove_empty, noop),
    ]



# Generated by Django 2.2.1 on 2019-05-18 10:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0014_auto_20190518_1046'),
    ]

    operations = [
        migrations.AlterField(
            model_name='certificate',
            name='revoked_reason',
            field=models.CharField(blank=True, choices=[('aa_compromise', 'Attribute Authority compromised'), ('affiliation_changed', 'Affiliation changed'), ('ca_compromise', 'CA compromised'), ('certificate_hold', 'On Hold'), ('cessation_of_operation', 'Cessation of operation'), ('key_compromise', 'Key compromised'), ('privilege_withdrawn', 'Privilege withdrawn'), ('remove_from_crl', 'Removed from CRL'), ('superseded', 'Superseded'), ('unspecified', 'Unspecified')], max_length=32, null=True, verbose_name='Reason for revokation'),
        ),
        migrations.AlterField(
            model_name='certificateauthority',
            name='revoked_reason',
            field=models.CharField(blank=True, choices=[('aa_compromise', 'Attribute Authority compromised'), ('affiliation_changed', 'Affiliation changed'), ('ca_compromise', 'CA compromised'), ('certificate_hold', 'On Hold'), ('cessation_of_operation', 'Cessation of operation'), ('key_compromise', 'Key compromised'), ('privilege_withdrawn', 'Privilege withdrawn'), ('remove_from_crl', 'Removed from CRL'), ('superseded', 'Superseded'), ('unspecified', 'Unspecified')], max_length=32, null=True, verbose_name='Reason for revokation'),
        ),
    ]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Functions to render extensions as text."""

import textwrap

from cryptography import x509
from cryptography.x509.oid import AuthorityInformationAccessOID

from django_ca import typehints
from django_ca.constants import EXTENDED_KEY_USAGE_NAMES
from django_ca.extensions.utils import key_usage_items, signed_certificate_timestamp_values
from django_ca.utils import bytes_to_hex, format_general_name, format_name


def _authority_information_access_as_text(value: typehints.InformationAccessExtensionType) -> str:
    lines = []
    issuers = [ad for ad in value if ad.access_method == AuthorityInformationAccessOID.CA_ISSUERS]
    ocsp = [ad for ad in value if ad.access_method == AuthorityInformationAccessOID.OCSP]
    if issuers:
        lines.append("CA Issuers:")
        lines += [f"  * {format_general_name(ad.access_location)}" for ad in issuers]
    if ocsp:
        lines.append("OCSP:")
        lines += [f"  * {format_general_name(ad.access_location)}" for ad in ocsp]
    return "\n".join(lines)


def _authority_key_identifier_as_text(value: x509.AuthorityKeyIdentifier) -> str:
    lines = []
    if value.key_identifier:
        lines.append(f"* KeyID: {bytes_to_hex(value.key_identifier)}")
    if value.authority_cert_issuer:
        lines.append("* Issuer:")
        lines += [f"  * {format_general_name(aci)}" for aci in value.authority_cert_issuer]
    if value.authority_cert_serial_number is not None:
        lines.append(f"* Serial: {value.authority_cert_serial_number}")
    return "\n".join(lines)


def _basic_constraints_as_text(value: x509.BasicConstraints) -> str:
    if value.ca is True:
        text = "CA:TRUE"
    else:
        text = "CA:FALSE"
    if value.path_length is not None:
        text += f", path length:{value.path_length}"

    return text


def _certificate_policies_as_text(value: x509.CertificatePolicies) -> str:
    lines = []

    # pylint: disable-next=too-many-nested-blocks
    for policy in value:
        lines.append(f"* Policy Identifier: {policy.policy_identifier.dotted_string}")

        if policy.policy_qualifiers:
            lines.append("  Policy Qualifiers:")
            for qualifier in policy.policy_qualifiers:
                if isinstance(qualifier, str):
                    lines += textwrap.wrap(qualifier, 76, initial_indent="  * ", subsequent_indent="    ")
                else:
                    lines.append("  * User Notice:")
                    if qualifier.explicit_text:
                        lines += textwrap.wrap(
                            f"Explicit Text: {qualifier.explicit_text}\n",
                            initial_indent="    * ",
                            subsequent_indent="        ",
                            width=76,
                        )
                    if qualifier.notice_reference:
                        lines.append("    * Notice Reference:")
                        if qualifier.notice_reference.organization:  # pragma: no branch
                            lines.append(f"      * Organization: {qualifier.notice_reference.organization}")
                        if qualifier.notice_reference.notice_numbers:
                            lines.append(
                                f"      * Notice Numbers: {qualifier.notice_reference.notice_numbers}"
                            )
        else:
            lines.append("  No Policy Qualifiers")
    return "\n".join(lines)


def _distribution_points_as_text(value: typehints.CRLExtensionType) -> str:
    lines = []
    for distribution_point in value:
        lines.append("* DistributionPoint:")

        if distribution_point.full_name:
            lines.append("  * Full Name:")
            lines += [f"    * {format_general_name(name)}" for name in distribution_point.full_name]
        elif distribution_point.relative_name:
            lines.append(f"  * Relative Name: {format_name(distribution_point.relative_name)}")
        else:  # pragma: no cover; either full_name or relative_name must be not-None.
            raise ValueError("Either full_name or relative_name must be not None.")

        if distribution_point.crl_issuer:
            lines.append("  * CRL Issuer:")
            lines += [f"    * {format_general_name(issuer)}" for issuer in distribution_point.crl_issuer]
        if distribution_point.reasons:
            reasons = ", ".join(sorted([r.name for r in distribution_point.reasons]))
            lines.append(f"  * Reasons: {reasons}")
    return "\n".join(lines)


def _key_usage_as_text(value: x509.KeyUsage) -> str:
    return "\n".join(f"* {name}" for name in sorted(key_usage_items(value)))


def _name_constraints_as_text(value: x509.NameConstraints) -> str:
    lines = []
    if value.permitted_subtrees:
        lines.append("Permitted:")
        lines += [f"  * {format_general_name(name)}" for name in value.permitted_subtrees]
    if value.excluded_subtrees:
        lines.append("Excluded:")
        lines += [f"  * {format_general_name(name)}" for name in value.excluded_subtrees]
    return "\n".join(lines)


def _policy_constraints_as_text(value: x509.PolicyConstraints) -> str:
    lines = []
    if value.inhibit_policy_mapping is not None:
        lines.append(f"* InhibitPolicyMapping: {value.inhibit_policy_mapping}")
    if value.require_explicit_policy is not None:
        lines.append(f"* RequireExplicitPolicy: {value.require_explicit_policy}")

    return "\n".join(lines)


def _signed_certificate_timestamps_as_text(value: typehints.SignedCertificateTimestampType) -> str:
    lines = []
    for sct in value:
        entry_type, version, log_id, timestamp = signed_certificate_timestamp_values(sct)

        lines += [
            f"* {entry_type} ({version}):",
            f"    Timestamp: {timestamp}",
            f"    Log ID: {log_id}",
        ]

    return "\n".join(lines)


def _tls_feature_as_text(value: x509.TLSFeature) -> str:
    lines = []
    for feature in value:
        if feature == x509.TLSFeatureType.status_request:
            lines.append("* status_request (OCSPMustStaple)")
        elif feature == x509.TLSFeatureType.status_request_v2:
            lines.append("* status_request_v2 (MultipleCertStatusRequest)")
        else:  # pragma: no cover
            # COVERAGE NOTE: we support all types, so this should never be raised. The descriptive error
            # message is just here in case a new thing ever comes up.
            raise ValueError(f"Unknown TLSFeatureType encountered: {feature}")
    return "\n".join(sorted(lines))


def extension_as_text(value: x509.ExtensionType) -> str:  # pylint: disable=too-many-return-statements
    """Return the given extension value as human-readable text."""
    if isinstance(value, (x509.OCSPNoCheck, x509.PrecertPoison)):
        return "Yes"  # no need for extra function
    if isinstance(value, (x509.FreshestCRL, x509.CRLDistributionPoints)):
        return _distribution_points_as_text(value)
    if isinstance(value, (x509.IssuerAlternativeName, x509.SubjectAlternativeName)):
        return "\n".join(f"* {format_general_name(name)}" for name in value)
    if isinstance(value, (x509.PrecertificateSignedCertificateTimestamps, x509.SignedCertificateTimestamps)):
        return _signed_certificate_timestamps_as_text(value)
    if isinstance(value, (x509.AuthorityInformationAccess, x509.SubjectInformationAccess)):
        return _authority_information_access_as_text(value)
    if isinstance(value, x509.AuthorityKeyIdentifier):
        return _authority_key_identifier_as_text(value)
    if isinstance(value, x509.BasicConstraints):
        return _basic_constraints_as_text(value)
    if isinstance(value, x509.CertificatePolicies):
        return _certificate_policies_as_text(value)
    if isinstance(value, x509.ExtendedKeyUsage):
        return "\n".join(sorted(f"* {EXTENDED_KEY_USAGE_NAMES[usage]}" for usage in value))
    if isinstance(value, x509.InhibitAnyPolicy):
        return str(value.skip_certs)
    if isinstance(value, x509.KeyUsage):
        return _key_usage_as_text(value)
    if isinstance(value, x509.NameConstraints):
        return _name_constraints_as_text(value)
    if isinstance(value, x509.PolicyConstraints):
        return _policy_constraints_as_text(value)
    if isinstance(value, x509.SubjectKeyIdentifier):
        return bytes_to_hex(value.key_identifier)
    if isinstance(value, x509.TLSFeature):
        return _tls_feature_as_text(value)
    if isinstance(value, x509.UnrecognizedExtension):
        return bytes_to_hex(value.value)
    if isinstance(value, x509.ExtensionType):
        raise TypeError(
            f"{value.__class__.__name__} (oid: {value.oid.dotted_string}): Unknown extension type."
        )
    raise TypeError(f"{value.__class__.__name__}: Not a cryptography.x509.ExtensionType.")



# Generated by Django 2.2.2 on 2019-07-06 15:48

from django.db import migrations


def rm_colons(apps, schema_editor):
    Certificate = apps.get_model('django_ca', 'Certificate')
    CertificateAuthority = apps.get_model('django_ca', 'CertificateAuthority')

    for ca in CertificateAuthority.objects.all():
        ca.serial = ca.serial.replace(':', '')
        ca.save()
    for cert in Certificate.objects.all():
        cert.serial = cert.serial.replace(':', '')
        cert.save()


def add_colons(apps, schema_editor):
    Certificate = apps.get_model('django_ca', 'Certificate')
    CertificateAuthority = apps.get_model('django_ca', 'CertificateAuthority')

    add_c = lambda s: ':'.join([s[i:i + 2] for i in range(0, len(s), 2)])  # NOQA

    for ca in CertificateAuthority.objects.all():
        ca.serial = add_c(ca.serial)
        ca.save()
    for cert in Certificate.objects.all():
        cert.serial = add_c(cert.serial)
        cert.save()


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0015_auto_20190518_1050'),
    ]

    operations = [
        migrations.RunPython(rm_colons, add_colons)
    ]



# Generated by Django 3.2 on 2021-05-01 12:58

from django.db import migrations
import django_ca.modelfields


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0025_auto_20210430_1132'),
    ]

    operations = [
        migrations.AlterField(
            model_name='certificate',
            name='pub',
            field=django_ca.modelfields.CertificateField(verbose_name='Public key'),
        ),
        migrations.AlterField(
            model_name='certificateauthority',
            name='pub',
            field=django_ca.modelfields.CertificateField(verbose_name='Public key'),
        ),
    ]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Asynchronous Celery tasks for django-ca.

.. seealso:: https://docs.celeryproject.org/en/stable/index.html
"""

import logging
import typing
from datetime import datetime, timedelta, timezone as tz
from http import HTTPStatus
from typing import Any, Iterable, List, Optional, Tuple

import requests

from cryptography import x509
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.x509.oid import ExtensionOID

from django.db import transaction
from django.utils import timezone

from django_ca import ca_settings, constants
from django_ca.acme.validation import validate_dns_01
from django_ca.constants import EXTENSION_DEFAULT_CRITICAL
from django_ca.models import (
    AcmeAuthorization,
    AcmeCertificate,
    AcmeChallenge,
    AcmeOrder,
    Certificate,
    CertificateAuthority,
)
from django_ca.profiles import profiles
from django_ca.typehints import AllowedHashTypes, ParsableKeyType
from django_ca.utils import parse_general_name

log = logging.getLogger(__name__)

FuncTypeVar = typing.TypeVar("FuncTypeVar", bound=typing.Callable[..., Any])

try:
    from celery import shared_task
    from celery.local import Proxy
except ImportError:

    def shared_task(func: FuncTypeVar) -> "Proxy[FuncTypeVar]":
        """Dummy decorator so that we can use the decorator whether celery is installed or not."""

        # We do not yet need this, but might come in handy in the future:
        # func.delay = lambda *a, **kw: func(*a, **kw)
        # func.apply_async = lambda *a, **kw: func(*a, **kw)
        func.delay = func  # type: ignore[attr-defined]
        return typing.cast("Proxy[FuncTypeVar]", func)


def run_task(task: "Proxy[FuncTypeVar]", *args: Any, **kwargs: Any) -> Any:
    """Function that passes `task` to celery or invokes it directly, depending on if Celery is installed."""
    eager = kwargs.pop("eager", False)

    if ca_settings.CA_USE_CELERY is True and eager is False:
        return task.delay(*args, **kwargs)

    return task(*args, **kwargs)


@shared_task
def cache_crl(serial: str, **kwargs: Any) -> None:
    """Task to cache the CRL for a given CA."""
    ca = CertificateAuthority.objects.get(serial=serial)
    ca.cache_crls(**kwargs)


@shared_task
def cache_crls(serials: Optional[Iterable[str]] = None) -> None:
    """Task to cache the CRLs for all CAs."""
    if serials is None:  # pragma: no cover; just to make mypy happy
        serials = []

    if not serials:
        serials = typing.cast(
            Iterable[str], CertificateAuthority.objects.usable().values_list("serial", flat=True)
        )

    for serial in serials:
        run_task(cache_crl, serial)


@shared_task
def generate_ocsp_key(
    serial: str,
    profile: str = "ocsp",
    expires: Optional[int] = None,
    algorithm: Optional[str] = None,
    password: Optional[str] = None,
    key_size: Optional[int] = None,
    key_type: Optional[ParsableKeyType] = None,
    elliptic_curve: Optional[str] = None,
    autogenerated: bool = True,
    force: bool = False,
) -> Optional[Tuple[str, str, int]]:
    """Task to generate an OCSP key for the CA named by `serial`.

    The `serial` names the certificate authority for which to regenerate the OCSP responder certificate. All
    other arguments are passed on to :py:func:`~django_ca.models.CertificateAuthority.generate_ocsp_key`.

    The task returns the private and public key paths and the *primary key* of the generated certificate if
    a new certificate was generated, otherwise it returns ``None``.
    """

    ca: CertificateAuthority = CertificateAuthority.objects.get(serial=serial)

    parsed_expires: Optional[timedelta] = None
    parsed_algorithm: Optional[AllowedHashTypes] = None
    parsed_curve: Optional[ec.EllipticCurve] = None
    if expires is not None:
        parsed_expires = timedelta(seconds=expires)
    if algorithm is not None:
        parsed_algorithm = constants.HASH_ALGORITHM_TYPES[algorithm]()
    if elliptic_curve is not None:
        parsed_curve = constants.ELLIPTIC_CURVE_TYPES[elliptic_curve]()

    value = ca.generate_ocsp_key(
        profile=profile,
        expires=parsed_expires,
        algorithm=parsed_algorithm,
        password=password,
        key_size=key_size,
        key_type=key_type,
        elliptic_curve=parsed_curve,
        autogenerated=autogenerated,
        force=force,
    )
    if value is not None:
        private_path, cert_path, cert = value
        return private_path, cert_path, cert.pk
    return None


@shared_task
def generate_ocsp_keys(**kwargs: Any) -> List[Optional[Tuple[str, str, int]]]:
    """Task to generate an OCSP keys for all usable CAs."""
    keys = []
    for serial in CertificateAuthority.objects.usable().values_list("serial", flat=True):
        keys.append(generate_ocsp_key(serial, **kwargs))
    return keys


@shared_task
@transaction.atomic
def acme_validate_challenge(challenge_pk: int) -> None:
    """Validate an ACME challenge."""
    if not ca_settings.CA_ENABLE_ACME:
        log.error("ACME is not enabled.")
        return

    try:
        challenge = AcmeChallenge.objects.url().get(pk=challenge_pk)
    except AcmeChallenge.DoesNotExist:
        log.error("Challenge with id=%s not found", challenge_pk)
        return

    # Whoever is invoking this task is responsible for setting the status to "processing" first.
    if challenge.status != AcmeChallenge.STATUS_PROCESSING:
        log.error(
            "%s: %s: Invalid state (must be %s)", challenge, challenge.status, AcmeChallenge.STATUS_PROCESSING
        )
        return

    # If the auth cannot be used for validation, neither can this challenge. We check auth.usable instead of
    # challenge.usable b/c a challenge in the "processing" state is not "usable" (= it is already being used).
    if challenge.auth.usable is False:
        log.error("%s: Authentication is not usable", challenge)
        return

    # General data for challenge validation
    value = challenge.auth.value

    # Challenge is marked as invalid by default
    challenge_valid = False

    # Validate HTTP challenge (only thing supported so far)
    if challenge.type == AcmeChallenge.TYPE_HTTP_01:
        decoded_token = challenge.encoded_token.decode("utf-8")
        expected = challenge.expected

        if requests is None:  # pragma: no cover
            log.error("requests is not installed, cannot do http-01 challenge validation.")
            return

        url = f"http://{value}/.well-known/acme-challenge/{decoded_token}"

        try:
            with requests.get(url, timeout=1, stream=True) as response:
                # Only fetch the response body if the status code is HTTP 200 (OK)
                if response.status_code == HTTPStatus.OK:
                    # Only fetch the expected number of bytes to prevent a large file ending up in memory
                    # But fetch one extra byte (if available) to make sure that response has no extra bytes
                    received = response.raw.read(len(expected) + 1, decode_content=True)
                    challenge_valid = received == expected
        except Exception as ex:  # pylint: disable=broad-except
            log.exception(ex)
    elif challenge.type == AcmeChallenge.TYPE_DNS_01:
        challenge_valid = validate_dns_01(challenge)

    # TODO: support ALPN_01 challenges
    # elif challenge.type == AcmeChallenge.TYPE_TLS_ALPN_01:
    #     host = socket.gethostbyname(value)
    #     sni_cert = crypto_util.probe_sni(
    #         host=host, port=443, name=value, alpn_protocols=[TlsAlpnProtocol.V1]
    #     )
    else:
        log.error("%s: Challenge type is not supported.", challenge)

    # Transition state of the challenge depending on if the challenge is valid or not. RFC8555, Section 7.1.6:
    #
    #   "If validation is successful, the challenge moves to the "valid" state; if there is an error, the
    #   challenge moves to the "invalid" state."
    #
    # We also transition the matching authorization object:
    #
    #   "If one of the challenges listed in the authorization transitions to the "valid" state, then the
    #   authorization also changes to the "valid" state.  If the client attempts to fulfill a challenge and
    #   fails, or if there is an error while the authorization is still pending, then the authorization
    #   transitions to the "invalid" state.
    #
    # We also transition the matching order object (section 7.4):
    #
    #   "* ready: The server agrees that the requirements have been fulfilled, and is awaiting finalization.
    #   Submit a finalization request."
    if challenge_valid:
        challenge.status = AcmeChallenge.STATUS_VALID
        challenge.validated = timezone.now()
        challenge.auth.status = AcmeAuthorization.STATUS_VALID

        # Set the order status to READY if all challenges are valid
        auths = AcmeAuthorization.objects.filter(order=challenge.auth.order)
        auths = auths.exclude(status=AcmeAuthorization.STATUS_VALID)
        if not auths.exclude(pk=challenge.auth.pk).exists():
            log.info("Order is now valid")
            challenge.auth.order.status = AcmeOrder.STATUS_READY
    else:
        challenge.status = AcmeChallenge.STATUS_INVALID

        # RFC 8555, section 7.1.6:
        #
        # If the client attempts to fulfill a challenge and fails, or if there is an error while the
        # authorization is still pending, then the authorization transitions to the "invalid" state.
        challenge.auth.status = AcmeAuthorization.STATUS_INVALID

        # RFC 8555, section 7.1.6:
        #
        #   If an error occurs at any of these stages, the order moves to the "invalid" state.
        challenge.auth.order.status = AcmeOrder.STATUS_INVALID

    log.info("%s is %s", challenge, challenge.status)
    challenge.save()
    challenge.auth.save()
    challenge.auth.order.save()


@shared_task
@transaction.atomic
def acme_issue_certificate(acme_certificate_pk: int) -> None:
    """Actually issue an ACME certificate."""
    if not ca_settings.CA_ENABLE_ACME:
        log.error("ACME is not enabled.")
        return

    try:
        acme_cert = AcmeCertificate.objects.select_related("order__account__ca").get(pk=acme_certificate_pk)
    except AcmeCertificate.DoesNotExist:
        log.error("Certificate with id=%s not found", acme_certificate_pk)
        return

    if acme_cert.usable is False:
        log.error("%s: Cannot issue certificate for this order", acme_cert.order)
        return

    names = [a.subject_alternative_name for a in acme_cert.order.authorizations.all()]
    log.info("%s: Issuing certificate for %s", acme_cert.order, ",".join(names))
    subject_alternative_names = x509.SubjectAlternativeName([parse_general_name(name) for name in names])

    extensions = [
        x509.Extension(
            oid=ExtensionOID.SUBJECT_ALTERNATIVE_NAME,
            critical=EXTENSION_DEFAULT_CRITICAL[ExtensionOID.SUBJECT_ALTERNATIVE_NAME],
            value=subject_alternative_names,
        )
    ]

    ca = acme_cert.order.account.ca
    profile = profiles[ca.acme_profile]

    # Honor not_after from the order if set
    if acme_cert.order.not_after:
        expires = acme_cert.order.not_after

        # Make sure expires_datetime is tz-aware, even if USE_TZ=False.
        if timezone.is_naive(expires):
            expires = timezone.make_aware(expires)
    else:
        expires = datetime.now(tz=tz.utc) + ca_settings.ACME_DEFAULT_CERT_VALIDITY

    csr = acme_cert.parse_csr()

    # Finally, actually create a certificate
    cert = Certificate.objects.create_cert(
        ca, csr=csr, profile=profile, expires=expires, extensions=extensions
    )

    acme_cert.cert = cert
    acme_cert.order.status = AcmeOrder.STATUS_VALID
    acme_cert.order.save()
    acme_cert.save()


@shared_task
@transaction.atomic
def acme_cleanup() -> None:
    """Cleanup expired ACME orders."""

    if not ca_settings.CA_ENABLE_ACME:
        # NOTE: Since this task does only cleanup, log message is only info.
        log.info("ACME is not enabled, not doing anything.")
        return

    # Delete orders that expired more than a day ago.
    threshold = timezone.now() - timedelta(days=1)
    AcmeOrder.objects.filter(expires__lt=threshold).delete()



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Management command to list all available certificate authorities.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

from typing import Any

from django.core.management.base import CommandParser

from django_ca.management.base import BaseCommand
from django_ca.models import CertificateAuthority
from django_ca.querysets import CertificateAuthorityQuerySet
from django_ca.utils import add_colons


class Command(BaseCommand):  # pylint: disable=missing-class-docstring
    help = "List available certificate authorities."

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "-t", "--tree", default=False, action="store_true", help="Output data in a tree view."
        )

    def qs(self, qs: CertificateAuthorityQuerySet) -> CertificateAuthorityQuerySet:
        """Order given queryset appropriately."""
        return qs.order_by("expires", "name")

    def list_ca(self, ca: CertificateAuthority, indent: str = "") -> None:
        """Output list line for a given CA."""

        text = f"{indent}{add_colons(ca.serial)} - {ca.name}"
        if ca.enabled is False:
            text += " (disabled)"

        self.stdout.write(text)

    def list_children(self, ca: CertificateAuthority, indent: str = "") -> None:
        """Output list lines for children of the given CA."""

        children = list(enumerate(self.qs(ca.children.all()), 1))
        for index, child in children:
            if index == len(children):  # last element
                self.list_ca(child, indent=indent + "└───")
            else:
                self.list_ca(child, indent=indent + "│───")

            children_left = len(children) - index
            if children_left:
                child_indent = indent + "│   "
            else:
                child_indent = indent + "    "

            self.list_children(child, child_indent)

    def handle(self, tree: bool, **options: Any) -> None:
        if tree:
            for ca in self.qs(CertificateAuthority.objects.filter(parent__isnull=True)):
                self.list_ca(ca)
                self.list_children(ca)
        else:
            for ca in self.qs(CertificateAuthority.objects.all()):
                self.list_ca(ca)



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Management command to revoke a certificate.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""
from datetime import datetime, timezone as tz
from typing import Any, Optional

from django.conf import settings
from django.core.management.base import CommandError, CommandParser
from django.utils import timezone

from django_ca.constants import ReasonFlags
from django_ca.management.actions import ReasonAction
from django_ca.management.base import BaseCommand
from django_ca.management.mixins import CertCommandMixin
from django_ca.models import Certificate


class Command(CertCommandMixin, BaseCommand):  # pylint: disable=missing-class-docstring
    allow_revoked = True
    help = "Revoke a certificate."

    def add_arguments(self, parser: CommandParser) -> None:
        super().add_arguments(parser)

        # Get a good example timestamp in for the help text
        example = datetime.now(tz=tz.utc).replace(microsecond=0, second=0).isoformat()

        group = parser.add_argument_group("Revocation information")
        group.add_argument("--reason", action=ReasonAction, help="An optional reason for revocation.")
        group.add_argument(
            "--compromised",
            metavar="TIMESTAMP",
            type=datetime.fromisoformat,
            help=f"When the certificate was compromised, as an ISO 8601 timestamp (example: {example}).",
        )

    def handle(
        self, cert: Certificate, reason: ReasonFlags, compromised: Optional[datetime], **options: Any
    ) -> None:
        if cert.revoked:
            raise CommandError(f"{cert.serial}: Certificate is already revoked.")

        # Make sure that the timestamp is tz-aware (makes processing easier)
        if compromised is not None and timezone.is_naive(compromised):
            raise CommandError(f"{compromised.isoformat()}: Timestamp requires a timezone.")

        # Make sure that the certificate was compromised in the past
        if compromised is not None and compromised > datetime.now(tz=tz.utc):
            raise CommandError(f"{compromised.isoformat()}: Timestamp must be in the past.")

        # If compromised is passed and USE_TZ=False, convert the timestamp to a tz-naive timestamp
        if compromised is not None and settings.USE_TZ is False:
            compromised = timezone.make_naive(compromised)

        cert.revoke(reason=reason, compromised=compromised)



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Management command to import a certificate authority.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

import argparse
import os
import typing
from typing import Any, List, Optional

from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import Encoding, PrivateFormat
from cryptography.x509.oid import ExtensionOID

from django.core.files.base import ContentFile
from django.core.management.base import CommandError, CommandParser

from django_ca import ca_settings
from django_ca.management.actions import PasswordAction
from django_ca.management.base import BaseCommand
from django_ca.management.mixins import CertificateAuthorityDetailMixin
from django_ca.models import CertificateAuthority
from django_ca.utils import ca_storage, format_general_name


class Command(CertificateAuthorityDetailMixin, BaseCommand):
    """Implement :command:`manage.py import_ca`."""

    help = """Import an existing certificate authority.

Note that the private key will be copied to the directory configured by the CA_DIR setting."""

    def add_arguments(self, parser: CommandParser) -> None:
        self.add_ca(
            parser,
            "--parent",
            help_text="Make the CA an intermediate CA of the named CA. By default, this is a new root CA.",
            no_default=True,
        )
        self.add_password(
            parser, help_text="Password used to encrypt the private key. Pass no argument to be prompted."
        )
        parser.add_argument(
            "--import-password",
            nargs="?",
            action=PasswordAction,
            metavar="PASSWORD",
            prompt="Password to import CA: ",
            help="Password for the private key.",
        )

        self.add_acme_group(parser)
        self.add_ocsp_group(parser)
        self.add_ca_args(parser)

        parser.add_argument("name", help="Human-readable name of the CA")
        parser.add_argument(
            "key", help="Path to the private key (PEM or DER format).", type=argparse.FileType("rb")
        )
        parser.add_argument(
            "pem", help="Path to the public key (PEM or DER format).", type=argparse.FileType("rb")
        )

    def handle(  # pylint: disable=too-many-locals,too-many-arguments
        self,
        name: str,
        key: typing.BinaryIO,
        pem: typing.BinaryIO,
        parent: Optional[CertificateAuthority],
        password: Optional[bytes],
        import_password: Optional[bytes],
        sign_ca_issuer: str,
        sign_crl_full_name: List[str],
        sign_issuer_alternative_name: Optional[x509.Extension[x509.IssuerAlternativeName]],
        sign_ocsp_responder: str,
        # Certificate Policies extension
        sign_certificate_policies: Optional[x509.CertificatePolicies],
        sign_certificate_policies_critical: bool,
        # OCSP responder configuration
        ocsp_responder_key_validity: Optional[int],
        ocsp_response_validity: Optional[int],
        **options: Any,
    ) -> None:
        if not os.path.exists(ca_settings.CA_DIR):
            try:
                os.makedirs(ca_settings.CA_DIR)
            except PermissionError as ex:
                pem.close()
                key.close()
                raise CommandError(
                    f"{ca_settings.CA_DIR}: Could not create CA_DIR: Permission denied."
                ) from ex
            # FileNotFoundError shouldn't happen, whole point of this block is to create it

        pem_data = pem.read()
        key_data = key.read()

        # close reader objects (otherwise we get a ResourceWarning)
        key.close()
        pem.close()

        if sign_issuer_alternative_name is None:
            issuer_alternative_name = ""
        else:
            issuer_alternative_name = format_general_name(sign_issuer_alternative_name.value[0])
        sign_certificate_policies_ext = None
        if sign_certificate_policies is not None:
            sign_certificate_policies_ext = x509.Extension(
                oid=ExtensionOID.CERTIFICATE_POLICIES,
                critical=sign_certificate_policies_critical,
                value=sign_certificate_policies,
            )

        ca = CertificateAuthority(
            name=name,
            parent=parent,
            issuer_url=sign_ca_issuer,
            issuer_alt_name=issuer_alternative_name,
            ocsp_url=sign_ocsp_responder,
            crl_url="\n".join(sign_crl_full_name),
            sign_certificate_policies=sign_certificate_policies_ext,
        )

        # Set OCSP responder options
        if ocsp_responder_key_validity is not None:
            ca.ocsp_responder_key_validity = ocsp_responder_key_validity
        if ocsp_response_validity is not None:
            ca.ocsp_response_validity = ocsp_response_validity

        # Set ACME options
        if ca_settings.CA_ENABLE_ACME:  # pragma: no branch; never False because parser throws error already
            for param in ["acme_enabled", "acme_registration", "acme_requires_contact"]:
                if options[param] is not None:
                    setattr(ca, param, options[param])

            if acme_profile := options["acme_profile"]:
                ca.acme_profile = acme_profile

        # load public key
        try:
            pem_loaded = x509.load_pem_x509_certificate(pem_data)
        except Exception:  # pylint: disable=broad-except
            try:
                pem_loaded = x509.load_der_x509_certificate(pem_data)
            except Exception as ex:
                raise CommandError("Unable to load public key.") from ex
        ca.update_certificate(pem_loaded)
        serial = ca.serial.replace(":", "")
        ca.private_key_path = ca_storage.generate_filename(f"{serial}.key")

        # load private key
        try:
            key_loaded = serialization.load_pem_private_key(key_data, import_password)
        except Exception:  # pylint: disable=broad-except
            try:
                key_loaded = serialization.load_der_private_key(key_data, import_password)
            except Exception as ex:
                raise CommandError("Unable to load private key.") from ex

        if password is None:
            encryption: serialization.KeySerializationEncryption = serialization.NoEncryption()
        else:
            encryption = serialization.BestAvailableEncryption(password)

        # write private key to file
        pem_as_bytes = key_loaded.private_bytes(
            encoding=Encoding.PEM, format=PrivateFormat.PKCS8, encryption_algorithm=encryption
        )

        try:
            ca_storage.save(ca.private_key_path, ContentFile(pem_as_bytes))
        except PermissionError as ex:
            raise CommandError(
                f"{ca.private_key_path}: Permission denied: Could not open file for writing"
            ) from ex

        # Only save CA to database if we loaded all data and copied private key
        ca.save()



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Pydantic schemas for extensions."""

from typing import List, Optional, Union

from ninja import Schema
from pydantic import Field, root_validator, validator

from cryptography import x509
from cryptography.hazmat._oid import ExtensionOID

from django_ca import constants
from django_ca.typehints import SerializedDistributionPoint

DATETIME_EXAMPLE = "2023-07-30T10:06:35Z"


class AuthorityInformationAccessValueSchema(Schema):
    """Schema for the Authority Information Access extension value."""

    issuers: Optional[List[str]] = Field(example=["URI:https://example.com/issuer"])
    ocsp: Optional[List[str]] = Field(example=["URI:https://example.com/ocsp"])


class AuthorityInformationAccessSchema(Schema):
    """Schema for the Authority Information Access extension.

    This extension is usually derived from the certificate authority signing the extension and not specified
    via the API. If given via the API, the `issuers` and `ocsp` fields will replace the respective field of
    the extension from the certificate authority.
    """

    critical: bool = constants.EXTENSION_DEFAULT_CRITICAL[ExtensionOID.AUTHORITY_INFORMATION_ACCESS]
    value: AuthorityInformationAccessValueSchema


class NoticeReferenceSchema(Schema):
    """Schema for a Notice Reference."""

    organization: Optional[str]
    notice_numbers: List[int]


class UserNoticeSchema(Schema):
    """Schema for a User Notice."""

    notice_reference: Optional[NoticeReferenceSchema]
    explicit_text: Optional[str]


class PolicySchema(Schema):
    """Schema for a certificate policy."""

    policy_identifier: str = Field(example="1.2.3.4")
    policy_qualifiers: Optional[List[Union[str, UserNoticeSchema]]]


class CertificatePoliciesSchema(Schema):
    """Schema for the Certificate Policies extension."""

    critical: bool = constants.EXTENSION_DEFAULT_CRITICAL[ExtensionOID.CERTIFICATE_POLICIES]
    value: List[PolicySchema]


class CRLDistributionPointSchema(Schema):
    """Schema for a CRL Distribution Point.

    Note that in practice, this usually is just a single `full_name` with a URL pointing to the CRL.
    """

    full_name: Optional[List[str]] = Field(example=["URI:http://crl.example.com"])
    relative_name: Optional[str]
    crl_issuer: Optional[List[str]] = Field(example=["URI:http://crl-issuers.example.com"])
    reasons: Optional[List[x509.ReasonFlags]] = Field(
        example=["unspecified", "superseded", "cessationOfOperation"]
    )

    @root_validator
    def check_full_or_relative_name(  # pylint: disable=no-self-argument  # -> pydantic
        cls, values: SerializedDistributionPoint
    ) -> SerializedDistributionPoint:
        """Validate that the distribution point has either a full_name OR a relative_name."""
        full_name = values.get("full_name")
        relative_name = values.get("relative_name")
        if full_name and relative_name:
            raise ValueError("Distribution point must contain either full_name OR relative_name.")
        if not full_name and not relative_name:
            raise ValueError("Distribution point must contain one of full_name OR relative_name.")
        return values


class CRLDistributionPointsSchema(Schema):
    """Schema for the CRL Distribution Points extension.

    This extension is usually derived from the certificate authority signing the extension and not specified
    via the API. If given via the API, it will replace any the extension from the certificate authority.
    """

    critical: bool = constants.EXTENSION_DEFAULT_CRITICAL[ExtensionOID.CRL_DISTRIBUTION_POINTS]
    value: List[CRLDistributionPointSchema]


class ExtendedKeyUsageSchema(Schema):
    """Schema for the Extended Key Usage extension."""

    critical: bool = constants.EXTENSION_DEFAULT_CRITICAL[ExtensionOID.EXTENDED_KEY_USAGE]
    value: List[str] = Field(example=["serverAuth", "clientAuth"])


class FreshestCRLSchema(Schema):
    """Schema for the Freshest CRL extension."""

    critical: bool = constants.EXTENSION_DEFAULT_CRITICAL[ExtensionOID.FRESHEST_CRL]
    value: List[CRLDistributionPointSchema]


class KeyUsageSchema(Schema):
    """Schema for the Key Usage extension."""

    critical: bool = constants.EXTENSION_DEFAULT_CRITICAL[ExtensionOID.KEY_USAGE]
    value: List[str] = Field(
        example=["digitalSignature", "keyEncipherment"], enum=sorted(constants.KEY_USAGE_NAMES.values())
    )

    @validator("value")
    def validate_key_usage(  # pylint: disable=no-self-argument  # -> pydantic
        cls, values: List[str]
    ) -> List[str]:
        """Make sure that only valid key usages are sent."""
        valid_values = tuple(constants.KEY_USAGE_NAMES.values())
        for key_usage in values:
            if key_usage not in valid_values:
                raise ValueError(f"{key_usage}: Invalid key usage.")
        return values


class OCSPNoCheckSchema(Schema):
    """Schema for the OCSP No Check extension."""

    critical: bool = constants.EXTENSION_DEFAULT_CRITICAL[ExtensionOID.OCSP_NO_CHECK]


class SubjectAlternativeNameSchema(Schema):
    """Schema for a Subject Alternative Name extension."""

    critical: bool = constants.EXTENSION_DEFAULT_CRITICAL[ExtensionOID.SUBJECT_ALTERNATIVE_NAME]
    value: List[str] = Field(example=["DNS:example.com", "IP:127.0.0.1"])


class TLSFeatureSchema(Schema):
    """Schema for the TLS Feature extension."""

    critical: bool = constants.EXTENSION_DEFAULT_CRITICAL[ExtensionOID.TLS_FEATURE]
    value: List[str] = Field(example=["OCSPMustStaple"])


class ExtensionsSchema(Schema):
    """Schema for all extensions that may be added via the API."""

    authority_information_access: Optional[AuthorityInformationAccessSchema]
    certificate_policies: Optional[CertificatePoliciesSchema]
    crl_distribution_points: Optional[CRLDistributionPointsSchema]
    extended_key_usage: Optional[ExtendedKeyUsageSchema]
    freshest_crl: Optional[CRLDistributionPointsSchema]
    key_usage: Optional[KeyUsageSchema]
    ocsp_no_check: Optional[OCSPNoCheckSchema]
    subject_alternative_name: Optional[SubjectAlternativeNameSchema]
    tls_feature: Optional[TLSFeatureSchema]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Views for the django-ca app.

.. seealso::

   * https://docs.djangoproject.com/en/dev/topics/class-based-views/
   * https://django-ca.readthedocs.io/en/latest/python/views.html
"""

import base64
import binascii
import logging
import os
import typing
from datetime import datetime, timedelta
from http import HTTPStatus
from typing import Any, Optional, Union

from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.types import CertificateIssuerPrivateKeyTypes
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.x509 import ExtensionNotFound, OCSPNonce, load_pem_x509_certificate, ocsp

from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest, HttpResponse, HttpResponseServerError
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import View
from django.views.generic.detail import SingleObjectMixin

from django_ca import ca_settings, constants
from django_ca.models import Certificate, CertificateAuthority
from django_ca.utils import SERIAL_RE, get_crl_cache_key, int_to_hex, parse_encoding, read_file

log = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from django.http.response import HttpResponseBase

    SingleObjectMixinBase = SingleObjectMixin[CertificateAuthority]
else:
    SingleObjectMixinBase = SingleObjectMixin


class CertificateRevocationListView(View, SingleObjectMixinBase):
    """Generic view that provides Certificate Revocation Lists (CRLs)."""

    slug_field = "serial"
    slug_url_kwarg = "serial"
    queryset = CertificateAuthority.objects.all().prefetch_related("certificate_set")

    password = None
    """Password used to load the private key of the certificate authority. If not set, the private key is
    assumed to be unencrypted."""

    # parameters for the CRL itself
    type = Encoding.DER
    """Encoding for CRL."""

    scope: Optional[typing.Literal["ca", "user", "attribute"]] = "user"
    """Set to ``"user"`` to limit CRL to certificates or ``"ca"`` to certificate authorities or ``None`` to
    include both."""

    expires = 600
    """CRL expires in this many seconds."""

    # header used in the request
    content_type = None
    """Value of the Content-Type header used in the response. For CRLs in PEM format, use ``text/plain``."""

    include_issuing_distribution_point: Optional[bool] = None
    """Boolean flag to force inclusion/exclusion of IssuingDistributionPoint extension."""

    def get(self, request: HttpRequest, serial: str) -> HttpResponse:
        # pylint: disable=missing-function-docstring; standard Django view function
        encoding = parse_encoding(request.GET.get("encoding", self.type))

        ca = self.get_object()

        cache_key = get_crl_cache_key(serial, encoding=encoding, scope=self.scope)

        crl = cache.get(cache_key)
        if crl is None:
            # Catch this case early so that we can give a better error message
            if self.include_issuing_distribution_point is True and ca.parent is None and self.scope is None:
                raise ValueError(
                    "Cannot add IssuingDistributionPoint extension to CRLs with no scope for root CAs."
                )

            encoding = parse_encoding(self.type)
            crl = ca.get_crl(
                expires=self.expires,
                password=self.password,
                scope=self.scope,
                include_issuing_distribution_point=self.include_issuing_distribution_point,
            )
            crl = crl.public_bytes(encoding)
            cache.set(cache_key, crl, self.expires)

        content_type = self.content_type
        if content_type is None:
            if self.type == Encoding.DER:
                content_type = "application/pkix-crl"
            elif self.type == Encoding.PEM:
                content_type = "text/plain"
            else:  # pragma: no cover
                # DER/PEM are all known encoding types, so this shouldn't happen
                return HttpResponseServerError()

        return HttpResponse(crl, content_type=content_type)


@method_decorator(csrf_exempt, name="dispatch")
class OCSPView(View):
    """View to provide an OCSP responder."""

    ca: str = ""
    """The name or serial of your Certificate Authority."""

    responder_key: str = ""
    """Private key used for signing OCSP responses. A relative path used by :ref:`CA_FILE_STORAGE
    <settings-ca-file-storage>`."""

    responder_cert: Union[x509.Certificate, str] = ""
    """Public key of the responder.

    This may either be:

    * A relative path used by :ref:`CA_FILE_STORAGE <settings-ca-file-storage>`
    * A serial of a certificate as stored in the database
    * The PEM of the certificate as string
    * A loaded :py:class:`~cg:cryptography.x509.Certificate`
    """

    expires = 600
    """Time in seconds that the responses remain valid. The default is 600 seconds or ten minutes."""

    ca_ocsp = False
    """If set to ``True``, validate child CAs instead."""

    def get(self, request: HttpRequest, data: str) -> HttpResponse:
        # pylint: disable=missing-function-docstring; standard Django view function
        try:
            decoded_data = base64.b64decode(data)
        except binascii.Error:
            return self.malformed_request()

        try:
            return self.process_ocsp_request(decoded_data)
        except Exception as e:  # pylint: disable=broad-except; we really need to catch everything here
            log.exception(e)
            return self.fail()

    def post(self, request: HttpRequest) -> HttpResponse:
        # pylint: disable=missing-function-docstring; standard Django view function
        try:
            return self.process_ocsp_request(request.body)
        except Exception as e:  # pylint: disable=broad-except; we really need to catch everything here
            log.exception(e)
            return self.fail()

    def fail(self, status: ocsp.OCSPResponseStatus = ocsp.OCSPResponseStatus.INTERNAL_ERROR) -> HttpResponse:
        """Generic method to return a failure response."""
        return self.http_response(
            ocsp.OCSPResponseBuilder.build_unsuccessful(status).public_bytes(Encoding.DER)
        )

    def get_responder_key(self) -> CertificateIssuerPrivateKeyTypes:
        """Get the private key used to sign OCSP responses."""
        key = self.get_responder_key_data()
        loaded_key = serialization.load_pem_private_key(key, None)

        # Check that the private key is of a supported type
        if not isinstance(loaded_key, constants.PRIVATE_KEY_TYPES):
            log.error("%s: Unsupported private key type.", type(loaded_key))
            raise ValueError(f"{type(loaded_key)}: Unsupported private key type.")

        return loaded_key

    def get_responder_key_data(self) -> bytes:
        """Read the file containing the private key used to sign OCSP responses."""
        return read_file(self.responder_key)

    def get_responder_cert(self) -> x509.Certificate:
        """Get the public key used to sign OCSP responses."""
        # User configured a loaded certificate
        if isinstance(self.responder_cert, x509.Certificate):
            return self.responder_cert

        if self.responder_cert.startswith("-----BEGIN CERTIFICATE-----\n"):
            responder_cert = self.responder_cert.encode("utf-8")
        elif SERIAL_RE.match(self.responder_cert):
            serial = self.responder_cert.replace(":", "")
            return Certificate.objects.get(serial=serial).pub.loaded
        else:
            if os.path.isabs(self.responder_cert):
                log.warning(
                    "%s: OCSP responder uses absolute path to certificate. Please see %s.",
                    self.responder_cert,
                    ca_settings.CA_FILE_STORAGE_URL,
                )
            responder_cert = read_file(self.responder_cert)

        return load_pem_x509_certificate(responder_cert)

    def get_ca(self) -> CertificateAuthority:
        """Get the certificate authority for the request."""
        return CertificateAuthority.objects.get_by_serial_or_cn(self.ca)

    def get_cert(self, ca: CertificateAuthority, serial: str) -> Union[Certificate, CertificateAuthority]:
        """Get the certificate that was requested in the OCSP request."""
        if self.ca_ocsp is True:
            return CertificateAuthority.objects.filter(parent=ca).get(serial=serial)

        return Certificate.objects.filter(ca=ca).get(serial=serial)

    def get_expires(self, now: datetime) -> datetime:
        """Get the timestamp when the OCSP response expires."""
        return now + timedelta(seconds=self.expires)

    def http_response(self, data: bytes, status: int = HTTPStatus.OK) -> HttpResponse:
        """Get an HTTP OCSP response with given status and data."""
        return HttpResponse(data, status=status, content_type="application/ocsp-response")

    def malformed_request(self) -> HttpResponse:
        """Get a response for a malformed request."""
        return self.fail(ocsp.OCSPResponseStatus.MALFORMED_REQUEST)

    def process_ocsp_request(self, data: bytes) -> HttpResponse:
        """Process OCSP request data."""

        try:
            ocsp_req = ocsp.load_der_ocsp_request(data)
        except Exception as e:  # pylint: disable=broad-except; we really need to catch everything here
            log.exception(e)
            return self.malformed_request()

        # Fail if there are any critical extensions that we do not understand
        for ext in ocsp_req.extensions:
            if ext.critical and not isinstance(ext.value, OCSPNonce):  # pragma: no cover
                # It seems impossible to get cryptography to create such a request, so it's not tested
                return self.malformed_request()

        # Get CA and certificate
        try:
            ca = self.get_ca()
        except CertificateAuthority.DoesNotExist:
            log.error("%s: Certificate Authority could not be found.", self.ca)
            return self.fail()

        cert_serial = int_to_hex(ocsp_req.serial_number)

        # NOINSPECTION NOTE: PyCharm wrongly things that second except is already covered by the first.
        # noinspection PyExceptClausesOrder
        try:
            cert = self.get_cert(ca, cert_serial)
        except CertificateAuthority.DoesNotExist:
            log.warning("%s: OCSP request for unknown CA received.", cert_serial)
            return self.fail()
        except Certificate.DoesNotExist:
            log.warning("%s: OCSP request for unknown cert received.", cert_serial)
            return self.fail()

        # get key/cert for OCSP responder
        try:
            responder_key = self.get_responder_key()
            responder_cert = self.get_responder_cert()
        except Exception:  # pylint: disable=broad-except; we really need to catch everything here
            log.error("Could not read responder key/cert.")
            return self.fail()

        # get the certificate status
        if cert.revoked:
            status = ocsp.OCSPCertStatus.REVOKED
        else:
            status = ocsp.OCSPCertStatus.GOOD

        now = datetime.utcnow()
        builder = ocsp.OCSPResponseBuilder()
        expires = self.get_expires(now)
        builder = builder.add_response(
            cert=cert.pub.loaded,
            issuer=ca.pub.loaded,
            # The algorithm used must be the same as in the request, or "openssl ocsp" won't be able to
            # determine the status
            algorithm=ocsp_req.hash_algorithm,
            cert_status=status,
            this_update=now,
            next_update=expires,
            revocation_time=cert.get_revocation_time(),
            revocation_reason=cert.get_revocation_reason(),
        ).responder_id(ocsp.OCSPResponderEncoding.HASH, responder_cert)

        # Add the responder cert to the response, necessary because we (so far) always use delegate
        # certificates
        builder = builder.certificates([responder_cert])

        # Add OCSP nonce if present
        try:
            nonce = ocsp_req.extensions.get_extension_for_class(OCSPNonce)
            builder = builder.add_extension(nonce.value, critical=nonce.critical)
        except ExtensionNotFound:
            pass

        response = builder.sign(responder_key, responder_cert.signature_hash_algorithm)
        return self.http_response(response.public_bytes(Encoding.DER))


@method_decorator(csrf_exempt, name="dispatch")
class GenericOCSPView(OCSPView):
    """View providing auto-configured OCSP functionality.

    This view assumes that ``ocsp/$ca_serial.(key|pem)`` point to the private/public key of a responder
    certificate as created by :py:class:`~django_ca.tasks.generate_ocsp_keys`. The ``serial`` URL keyword
    argument must be the serial for this CA.
    """

    auto_ca: CertificateAuthority

    # NOINSPECTION/TYPE NOTE: It's okay to be more specific here
    # noinspection PyMethodOverriding
    def dispatch(  # type: ignore[override]
        self, request: HttpRequest, serial: str, **kwargs: Any
    ) -> "HttpResponseBase":
        if request.method == "GET" and "data" not in kwargs:
            return self.http_method_not_allowed(request, serial, **kwargs)
        if request.method == "POST" and "data" in kwargs:
            return self.http_method_not_allowed(request, serial, **kwargs)

        # COVERAGE NOTE: Checking just for safety here.
        if not isinstance(serial, str):  # pragma: no cover
            raise ImproperlyConfigured("View expects a str for a serial")

        self.auto_ca = CertificateAuthority.objects.get(serial=serial)
        return super().dispatch(request, **kwargs)

    def get_ca(self) -> CertificateAuthority:
        return self.auto_ca

    def get_expires(self, now: datetime) -> datetime:
        return now + timedelta(seconds=self.auto_ca.ocsp_response_validity)

    def get_responder_key_data(self) -> bytes:
        serial = self.auto_ca.serial.replace(":", "")
        return read_file(f"ocsp/{serial}.key")

    def get_responder_cert(self) -> x509.Certificate:
        return self.auto_ca.ocsp_responder_certificate


class GenericCAIssuersView(View):
    """Generic view that returns a CA public key in DER format.

    This view serves the URL named in the ``issuers`` key in the
    :py:class:`~cg:cryptography.x509.AuthorityInformationAccess` extension.
    """

    def get(self, request: HttpRequest, serial: str) -> HttpResponse:
        # pylint: disable=missing-function-docstring; standard Django view function
        ca = CertificateAuthority.objects.get(serial=serial)
        return HttpResponse(ca.pub.der, content_type="application/pkix-cert")



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Management command to edit a certificate authority.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

from typing import Any, List, Optional

from cryptography import x509
from cryptography.x509.oid import ExtensionOID

from django.core.management.base import CommandError, CommandParser

from django_ca import ca_settings
from django_ca.management.base import BaseCommand
from django_ca.management.mixins import CertificateAuthorityDetailMixin
from django_ca.models import CertificateAuthority
from django_ca.utils import format_general_name


class Command(CertificateAuthorityDetailMixin, BaseCommand):
    """Implement :command:`manage.py edit_ca`."""

    help = "Edit a certificate authority."

    def add_arguments(self, parser: CommandParser) -> None:
        self.add_general_args(parser, default=None)
        self.add_ca(parser, "ca", allow_disabled=True)
        self.add_acme_group(parser)
        self.add_ocsp_group(parser)
        self.add_ca_args(parser)

        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "--enable",
            action="store_true",
            dest="enabled",
            default=None,
            help="Enable the certificate authority.",
        )
        group.add_argument(
            "--disable", action="store_false", dest="enabled", help="Disable the certificate authority."
        )

    def handle(  # pylint: disable=too-many-arguments
        self,
        ca: CertificateAuthority,
        sign_ca_issuer: str,
        sign_crl_full_name: List[str],
        sign_issuer_alternative_name: Optional[x509.Extension[x509.IssuerAlternativeName]],
        sign_ocsp_responder: str,
        enabled: Optional[bool],
        # Certificate Policies extension
        sign_certificate_policies: Optional[x509.CertificatePolicies],
        sign_certificate_policies_critical: bool,
        # OCSP responder configuration
        ocsp_responder_key_validity: Optional[int],
        ocsp_response_validity: Optional[int],
        **options: Any,
    ) -> None:
        if sign_ca_issuer:
            ca.issuer_url = sign_ca_issuer
        if sign_issuer_alternative_name:
            ca.issuer_alt_name = ",".join(
                [format_general_name(name) for name in sign_issuer_alternative_name.value]
            )
        if sign_ocsp_responder:
            ca.ocsp_url = sign_ocsp_responder
        if sign_crl_full_name:
            ca.crl_url = "\n".join(sign_crl_full_name)
        if sign_certificate_policies is not None:
            ca.sign_certificate_policies = x509.Extension(
                oid=ExtensionOID.CERTIFICATE_POLICIES,
                critical=sign_certificate_policies_critical,
                value=sign_certificate_policies,
            )

        if enabled is not None:
            ca.enabled = enabled

        if options["caa"] is not None:
            ca.caa_identity = options["caa"]
        if options["website"] is not None:
            ca.website = options["website"]
        if options["tos"] is not None:
            ca.terms_of_service = options["tos"]

        # Set ACME options
        if ca_settings.CA_ENABLE_ACME:  # pragma: no branch; never False because parser throws error already
            for param in ["acme_enabled", "acme_registration", "acme_requires_contact"]:
                if options[param] is not None:
                    setattr(ca, param, options[param])

            if acme_profile := options["acme_profile"]:
                if acme_profile not in ca_settings.CA_PROFILES:
                    raise CommandError(f"{acme_profile}: Profile is not defined.")
                ca.acme_profile = acme_profile

        # Set OCSP responder options
        if ocsp_responder_key_validity is not None:
            ca.ocsp_responder_key_validity = ocsp_responder_key_validity
        if ocsp_response_validity is not None:
            ca.ocsp_response_validity = ocsp_response_validity

        ca.save()



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Management command to resign an existing certificate.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

from datetime import timedelta
from typing import Any, List, Optional

from cryptography import x509
from cryptography.x509.oid import ExtensionOID, NameOID

from django.core.management.base import CommandError, CommandParser

from django_ca import ca_settings, constants
from django_ca.management.actions import CertificateAction
from django_ca.management.base import BaseSignCertCommand
from django_ca.models import Certificate, CertificateAuthority, Watcher
from django_ca.profiles import Profile, profiles
from django_ca.typehints import AllowedHashTypes, ExtensionMapping


class Command(BaseSignCertCommand):  # pylint: disable=missing-class-docstring
    help = f"""Sign a CSR and output signed certificate. The defaults depend on the configured
default profile, currently {ca_settings.CA_DEFAULT_PROFILE}."""

    add_extensions_help = "Override certificate extensions."
    subject_help = "Override subject for new certificate."

    def add_arguments(self, parser: CommandParser) -> None:
        self.add_base_args(parser, no_default_ca=True)
        self.add_profile(parser, """Use given profile to determine certificate expiry.""")
        parser.add_argument(
            "cert", action=CertificateAction, allow_revoked=True, help="The certificate to resign."
        )

    def get_profile(self, profile: Optional[str], cert: Certificate) -> Profile:
        """Get requested profile based on command line and given certificate."""
        if profile is not None:
            return profiles[profile]
        if cert.profile == "":
            return profiles[None]

        try:
            return profiles[cert.profile]
        except KeyError:
            # Occurs if the certificate specifies a profile which has since been removed from settings
            raise CommandError(  # pylint: disable=raise-missing-from
                f'Profile "{cert.profile}" for original certificate is no longer defined, please set one via the command line.'  # NOQA: E501
            )

    def handle(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        cert: Certificate,
        ca: Optional[CertificateAuthority],
        subject: Optional[x509.Name],
        expires: Optional[timedelta],
        watch: List[str],
        password: Optional[bytes],
        profile: Optional[str],
        algorithm: Optional[AllowedHashTypes],
        # Authority Information Access extension
        authority_information_access: x509.AuthorityInformationAccess,
        # Certificate Policies extension
        certificate_policies: Optional[x509.CertificatePolicies],
        certificate_policies_critical: bool,
        # CRL Distribution Points extension
        crl_full_names: Optional[List[x509.GeneralName]],
        crl_distribution_points_critical: bool,
        # Extended Key Usage extension
        extended_key_usage: Optional[x509.ExtendedKeyUsage],
        extended_key_usage_critical: bool,
        # Issuer Alternative Name extension:
        issuer_alternative_name: Optional[x509.IssuerAlternativeName],
        # Key Usage extension
        key_usage: Optional[x509.KeyUsage],
        key_usage_critical: bool,
        # OCSP No Check extension
        ocsp_no_check: bool,
        ocsp_no_check_critical: bool,
        # Subject Alternative Name extension
        subject_alternative_name: Optional[x509.SubjectAlternativeName],
        subject_alternative_name_critical: bool,
        # TLSFeature extension
        tls_feature: Optional[x509.TLSFeature],
        tls_feature_critical: bool,
        **options: Any,
    ) -> None:
        if ca is None:
            ca = cert.ca

        profile_obj = self.get_profile(profile, cert)
        self.test_options(ca=ca, password=password, expires=expires, profile=profile_obj, **options)

        # Get/validate signature hash algorithm
        algorithm = self.get_hash_algorithm(ca.key_type, algorithm, ca.algorithm)

        # get list of watchers
        if watch:
            watchers = [Watcher.from_addr(addr) for addr in watch]
        else:
            watchers = list(cert.watchers.all())

        if subject is None:
            subject = cert.subject

        # Process any extensions given via the command-line
        extensions: ExtensionMapping = {}

        if authority_information_access is not None:
            self._add_extension(
                extensions,
                authority_information_access,
                constants.EXTENSION_DEFAULT_CRITICAL[ExtensionOID.AUTHORITY_INFORMATION_ACCESS],
            )
        if certificate_policies is not None:
            self._add_extension(extensions, certificate_policies, certificate_policies_critical)
        if crl_full_names is not None:
            distribution_point = x509.DistributionPoint(
                full_name=crl_full_names, relative_name=None, crl_issuer=None, reasons=None
            )
            self._add_extension(
                extensions, x509.CRLDistributionPoints([distribution_point]), crl_distribution_points_critical
            )
        if extended_key_usage is not None:
            self._add_extension(extensions, extended_key_usage, extended_key_usage_critical)
        if issuer_alternative_name is not None:
            self._add_extension(
                extensions,
                issuer_alternative_name,
                constants.EXTENSION_DEFAULT_CRITICAL[ExtensionOID.ISSUER_ALTERNATIVE_NAME],
            )
        if key_usage is not None:
            self._add_extension(extensions, key_usage, key_usage_critical)
        if ocsp_no_check is True:
            self._add_extension(extensions, x509.OCSPNoCheck(), ocsp_no_check_critical)
        if subject_alternative_name is not None:
            self._add_extension(extensions, subject_alternative_name, subject_alternative_name_critical)
        if tls_feature is not None:
            self._add_extension(extensions, tls_feature, tls_feature_critical)

        # Copy over extensions from the original certificate (if not passed via the command-line)
        for oid, extension in cert.x509_extensions.items():
            # These extensions are handled by the manager itself based on the CA:
            if oid in (
                ExtensionOID.AUTHORITY_INFORMATION_ACCESS,
                ExtensionOID.AUTHORITY_KEY_IDENTIFIER,
                ExtensionOID.CRL_DISTRIBUTION_POINTS,
            ):
                continue

            # Add extensions not already added via the command line
            if oid not in extensions:
                extensions[oid] = extension

        # Verify that we have either a Common Name in the subject or a Subject Alternative Name extension
        # NOTE: This can only happen here in two edge cases:
        #   * Pass a subject without common name AND a certificate does *not* have a subject alternative name.
        #   * An imported certificate that has neither Common Name nor subject alternative name.
        common_names = subject.get_attributes_for_oid(NameOID.COMMON_NAME)
        if not common_names and ExtensionOID.SUBJECT_ALTERNATIVE_NAME not in extensions:
            raise CommandError(
                "Must give at least a Common Name in --subject or one or more "
                "--subject-alternative-name/--name arguments."
            )

        try:
            cert = Certificate.objects.create_cert(
                ca=ca,
                csr=cert.csr.loaded,
                profile=profile_obj,
                expires=expires,
                subject=subject,
                algorithm=algorithm,
                extensions=extensions.values(),
                password=password,
                cn_in_san=False,  # we already copy the SAN/CN from the original cert
            )
        except Exception as ex:
            raise CommandError(ex) from ex

        cert.watchers.add(*watchers)

        if options["out"]:
            with open(options["out"], "w", encoding="ascii") as stream:
                stream.write(cert.pub.pem)
        else:
            self.stdout.write(cert.pub.pem)



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""django-ca root module."""

from typing import Any, Tuple, Union

VersionTuple = Union[Tuple[int, int, int], Tuple[int, int, int, str, int]]

# WARNING: This module MUST NOT include any dependencies, as it is read by setup.py

# Between releases: (Major, Minor, Patch, "dev", 1)
# On a release: (Major, Minor, Patch)
# https://www.python.org/dev/peps/pep-0440/
# https://www.python.org/dev/peps/pep-0396/
# https://www.python.org/dev/peps/pep-0386/
VERSION: VersionTuple = (1, 26, 0)

# __version__ specified in PEP 0396, but we use the PEP 0440 format instead
__version__ = ".".join([str(e) for e in VERSION[:3]])
if len(VERSION) > 3:  # pragma: no cover
    # NOTE: dev_elements hack here only to make mypy happy in both dev and non-dev versions
    dev_elements: Tuple[Any, ...] = VERSION[3:5]
    __version__ += f".{''.join(str(e) for e in dev_elements)}"



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Django form fields related to django-ca."""

import abc
import typing
from typing import Any, Dict, Iterable, List, Optional, Sequence, Set, Tuple, Type, Union

from cryptography import x509
from cryptography.x509.oid import AuthorityInformationAccessOID

from django import forms
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from django_ca import ca_settings, constants, utils, widgets
from django_ca.constants import (
    EXTENDED_KEY_USAGE_HUMAN_READABLE_NAMES,
    EXTENDED_KEY_USAGE_NAMES,
    KEY_USAGE_NAMES,
    REVOCATION_REASONS,
)
from django_ca.extensions import get_extension_name
from django_ca.typehints import CRLExtensionTypeTypeVar, ExtensionTypeTypeVar
from django_ca.utils import parse_general_name
from django_ca.widgets import KeyValueWidget, SubjectWidget

if typing.TYPE_CHECKING:
    from django_ca.modelfields import LazyCertificateSigningRequest

_EXTENDED_KEY_USAGE_CHOICES = sorted(
    [(EXTENDED_KEY_USAGE_NAMES[oid], name) for oid, name in EXTENDED_KEY_USAGE_HUMAN_READABLE_NAMES.items()],
    key=lambda t: t[1],
)
_EXTENDED_KEY_USAGE_MAPPING = {serialized: oid for oid, serialized in EXTENDED_KEY_USAGE_NAMES.items()}


class CertificateSigningRequestField(forms.CharField):
    """A form field for `~cg:cryptography.x509.CertificateSigningRequest` encoded as PEM."""

    start = "-----BEGIN CERTIFICATE REQUEST-----"
    end = "-----END CERTIFICATE REQUEST-----"
    simple_validation_error = _(
        "Could not parse PEM-encoded CSR. They usually look like this: <pre>%(start)s\n...\n%(end)s</pre>"
    ) % {"start": start, "end": end}

    def __init__(self, **kwargs: Any) -> None:
        # COVERAGE NOTE: Below condition is never false, as we never pass a custom help text.
        if not kwargs.get("help_text"):  # pragma: no branch
            kwargs["help_text"] = _(
                """The Certificate Signing Request (CSR) in PEM format. To create a new one:
<span class="shell">openssl genrsa -out priv.pem 4096
openssl req -new -key priv.pem -out csr.pem -utf8 -batch -subj '/CN=example.com'
</span>"""
            )
        if not kwargs.get("widget"):  # pragma: no branch # we never pass a custom widget
            kwargs["widget"] = forms.Textarea
        super().__init__(**kwargs)
        self.widget.attrs.update({"cols": "64"})

    def prepare_value(self, value: Optional[Union[str, "LazyCertificateSigningRequest"]]) -> str:
        """Prepare a value to a form that can be shown in HTML.

        Unfortunately this function is not documented by Django at all but is called when a form is rendered.

        This function receives ``None`` when viewing an initial form for a new instance. If form validation
        fails, it receives the string as posted by the user (which might be an invalid string). When viewing
        an existing certificate, it will receive the LazyField instance of the object.
        """
        if value is None:  # for new objects
            return ""
        if isinstance(value, str):  # when form validation fails
            return value

        # COVERAGE NOTE: This would happen if the field is editable, but is always read-only.
        return value.pem  # pragma: no cover

    # TYPE NOTE: django-stubs typehints this as Optional[Any], but we only ever observed receiving ``str``.
    def to_python(self, value: str) -> x509.CertificateSigningRequest:  # type: ignore[override]
        """Coerce given str to correct data type, raises ValidationError if not possible.

        This function is called during form validation.
        """
        if not value.startswith(self.start) or not value.strip().endswith(self.end):
            raise forms.ValidationError(mark_safe(self.simple_validation_error))
        try:
            return x509.load_pem_x509_csr(value.encode("utf-8"))
        except ValueError as ex:
            raise forms.ValidationError(str(ex)) from ex


class ObjectIdentifierField(forms.CharField):
    """A form field for a :py:class:`~cg:cryptography.x509.ObjectIdentifier`."""

    default_error_messages = {"invalid-oid": _("%(value)s: The given OID is invalid.")}

    def to_python(self, value: str) -> Optional[x509.ObjectIdentifier]:  # type: ignore[override]
        if not value:
            return None

        try:
            return x509.ObjectIdentifier(value)
        except ValueError as ex:
            raise forms.ValidationError(
                self.error_messages["invalid-oid"], code="invalid-oid", params={"value": value}
            ) from ex


class KeyValueField(forms.MultiValueField):
    """Dynamic Key/Value field."""

    widget_class = KeyValueWidget

    def __init__(self, key_choices: Sequence[Tuple[str, str]], **kwargs: Any) -> None:
        fields = [
            forms.JSONField(required=True),
            # These two just serve as template for key/value rows
            forms.ChoiceField(choices=key_choices, required=False),
            forms.CharField(required=False),
        ]

        kwargs["require_all_fields"] = False  # ... or we'd require values in the template

        super().__init__(fields, **kwargs)
        self.widget = self.widget_class(key_choices=key_choices, attrs={"class": "key-value-input"})

    def compress(
        self, data_list: Tuple[List[Dict[str, str]], str, str]
    ) -> Optional[List[Dict[str, str]]]:  # pragma: no cover  # This class is never used directly so far
        if not data_list or not data_list[0]:
            return None
        return data_list[0]


class SubjectField(KeyValueField):
    """Specialized version of KeyValue field for a certificate subject."""

    widget_class = SubjectWidget

    def __init__(self, **kwargs: Any) -> None:
        key_choices = [(oid.dotted_string, name) for oid, name in constants.NAME_OID_DISPLAY_NAMES.items()]
        super().__init__(key_choices=key_choices, **kwargs)

    def compress(  # type: ignore[override]
        self, data_list: Tuple[List[Dict[str, str]], str, str]
    ) -> x509.Name:
        # Empty list happens when you press submit on a completely empty form
        if not data_list or not data_list[0]:
            return x509.Name([])

        values = data_list[0]
        errors = []
        attributes: List[x509.NameAttribute] = []
        found_oids: Set[x509.ObjectIdentifier] = set()
        for oid_dict in values:
            try:
                oid = x509.ObjectIdentifier(oid_dict["key"])

                # Check for duplicate OIDs that should not occur more than once
                if oid in found_oids and oid not in utils.MULTIPLE_OIDS:
                    oid_name = constants.NAME_OID_DISPLAY_NAMES[oid]
                    errors.append(_("%(attr)s: Attribute cannot occur more then once.") % {"attr": oid_name})
                else:
                    found_oids.add(oid)

                attr = x509.NameAttribute(oid=oid, value=oid_dict["value"])
            except Exception as ex:  # pylint: disable=broad-exception-caught  # docs don't specify class
                # Creating the NameAttribute failed (e.g. a country code that does *not* have two characters)
                errors.append(str(ex))
            else:
                attributes.append(attr)

        if errors:
            raise forms.ValidationError(errors)

        name = x509.Name(attributes)
        return name


class GeneralNamesField(forms.CharField):
    """MultipleChoice field for :py:class:`~cg:cryptography.x509.RelativeDistinguishedName`."""

    widget = widgets.GeneralNamesWidget
    default_error_messages = {
        "invalid": _("Unparsable General Name: %(error)s"),
    }

    def to_python(  # type: ignore[override]  # superclass uses Any for str, violates inheritance (in theory)
        self, value: str
    ) -> Optional[List[x509.GeneralName]]:
        if not value:
            return None

        values = []
        for line in value.splitlines():
            line = line.strip()
            if not line:
                continue

            try:
                values.append(parse_general_name(line))
            except ValueError as ex:
                raise forms.ValidationError(
                    self.error_messages["invalid"], params={"error": ex}, code="invalid"
                )
        if not values:
            return None

        return values


class RelativeDistinguishedNameField(forms.CharField):
    """MultipleChoice field for :py:class:`~cg:cryptography.x509.RelativeDistinguishedName`."""

    def to_python(  # type: ignore[override]  # superclass uses Any for str, violates inheritance (in theory)
        self, value: str
    ) -> Optional[x509.RelativeDistinguishedName]:
        if not value:
            return None

        rdns = x509.Name.from_rfc4514_string(value).rdns
        attributes = [attr for rdn in rdns for attr in rdn]
        return x509.RelativeDistinguishedName(attributes=attributes)


class ReasonsField(forms.MultipleChoiceField):
    """MultipleChoice field for :py:class:`~cg:cryptography.x509.ReasonFlags`.

    .. note::

       This field does NOT convert to x509.ReasonFlags itself but uses string values instead. The choice
       field always returns invalid choice errors otherwise.
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(choices=REVOCATION_REASONS, **kwargs)


class ExtensionField(forms.MultiValueField, typing.Generic[ExtensionTypeTypeVar], metaclass=abc.ABCMeta):
    """Base class for form fields that serialize to a :py:class:`~cg:cryptography.Extension`."""

    extension_type: Type[ExtensionTypeTypeVar]
    # TYPEHINT NOTE: the value can be handled by the get_fields() method.
    fields: Optional[Tuple[forms.Field, ...]] = None  # type: ignore[assignment]

    def __init__(self, **kwargs: Any) -> None:
        fields = self.get_fields() + (forms.BooleanField(required=False, initial=True),)
        kwargs.setdefault("label", get_extension_name(self.extension_type.oid))
        super().__init__(fields=fields, require_all_fields=False, **kwargs)

    def compress(self, data_list: List[Any]) -> Optional[x509.Extension[ExtensionTypeTypeVar]]:
        if not data_list:
            return None

        *value, critical = data_list
        ext_value = self.get_value(*value)
        if ext_value is None:
            return None
        return x509.Extension(critical=critical, oid=self.extension_type.oid, value=ext_value)

    def get_fields(self) -> Tuple[forms.Field, ...]:
        """Get the form fields used for this extension.

        Note that the `critical` input field is automatically appended.
        """
        if self.fields is not None:  # pragma: no branch
            return self.fields
        raise ValueError(  # pragma: no cover
            "ExtensionField must either set fields or implement get_fields()."
        )

    @abc.abstractmethod
    def get_value(self, *value: Any) -> Optional[ExtensionTypeTypeVar]:
        """Get the extension value from the "compressed" form representation.

        Return `None` if no value was set and the extension should **not** be added.
        """


class MultipleChoiceExtensionField(ExtensionField[ExtensionTypeTypeVar]):
    """Base class for extensions that are basically a multiple choice field (plus critical)."""

    choices = Tuple[Tuple[str, str], ...]
    widget: Type[widgets.MultipleChoiceExtensionWidget]

    def __init__(self, **kwargs: Any) -> None:
        kwargs["widget"] = self.widget(choices=self.choices)
        super().__init__(**kwargs)

    def get_fields(self) -> Tuple[forms.MultipleChoiceField]:
        return (forms.MultipleChoiceField(choices=self.choices, required=False),)

    def get_value(self, value: List[str]) -> Optional[ExtensionTypeTypeVar]:  # type: ignore[override]
        if not value:
            return None
        return self.get_values(value)

    @abc.abstractmethod
    def get_values(self, value: List[str]) -> Optional[ExtensionTypeTypeVar]:
        """Get the ExtensionType instance from the selected values."""


class DistributionPointField(ExtensionField[CRLExtensionTypeTypeVar]):
    """Base class for extensions with DistributionPoints."""

    default_error_messages = {
        "full-and-relative-name": _("You cannot provide both full_name and relative_name."),
        "no-dp-or-issuer": _("A DistributionPoint needs at least a full or relative name or a crl issuer."),
    }
    fields = (
        GeneralNamesField(required=False),  # full_name
        RelativeDistinguishedNameField(required=False),  # relative_name
        GeneralNamesField(required=False),  # crl_issuer
        ReasonsField(required=False),  # reasons
    )

    def get_value(  # type: ignore[override]
        self,
        full_name: List[x509.GeneralName],
        relative_distinguished_name: Optional[x509.RelativeDistinguishedName],
        crl_issuer: List[x509.GeneralName],
        reasons: Optional[Iterable[str]],
    ) -> Optional[CRLExtensionTypeTypeVar]:
        if not full_name:
            # TYPEHINT NOTE: Field returns empty list, which x509.DistributionPoint() treats different from
            #   None. Any other solution is less efficient, so we don't use them just for mypy.
            full_name = None  # type: ignore[assignment]
        if not crl_issuer:
            crl_issuer = None  # type: ignore[assignment]  # same as above for full_name

        if reasons:
            parsed_reasons = frozenset(x509.ReasonFlags[flag] for flag in reasons)
        else:
            parsed_reasons = None

        if full_name and relative_distinguished_name:
            raise forms.ValidationError(
                self.error_messages["full-and-relative-name"], code="full-and-relative-name"
            )

        if not full_name and not relative_distinguished_name and not crl_issuer:
            if reasons:
                # NOTE: cryptography does not yet validate this on its own:
                #   https://github.com/pyca/cryptography/pull/7710
                raise forms.ValidationError(self.error_messages["no-dp-or-issuer"], code="no-dp-or-issuer")
            return None  # nothing was entered at all

        distribution_point = x509.DistributionPoint(
            full_name=full_name,
            relative_name=relative_distinguished_name,
            crl_issuer=crl_issuer,
            reasons=parsed_reasons,
        )
        return self.extension_type(distribution_points=[distribution_point])


class AuthorityInformationAccessField(ExtensionField[x509.AuthorityInformationAccess]):
    """Form field for a :py:class:`~cg:cryptography.x509.AuthorityInformationAccess` extension."""

    extension_type = x509.AuthorityInformationAccess
    fields = (GeneralNamesField(required=False), GeneralNamesField(required=False))
    widget = widgets.AuthorityInformationAccessWidget

    def get_value(  # type: ignore[override]
        self, ca_issuers: List[x509.GeneralName], ocsp: List[x509.GeneralName]
    ) -> Optional[x509.AuthorityInformationAccess]:
        if not ca_issuers and not ocsp:
            return None
        descriptions = []
        if ocsp:
            descriptions += [
                x509.AccessDescription(access_method=AuthorityInformationAccessOID.OCSP, access_location=name)
                for name in ocsp
            ]
        if ca_issuers:
            descriptions += [
                x509.AccessDescription(
                    access_method=AuthorityInformationAccessOID.CA_ISSUERS, access_location=name
                )
                for name in ca_issuers
            ]
        return x509.AuthorityInformationAccess(descriptions=descriptions)


class CertificatePoliciesField(ExtensionField[x509.CertificatePolicies]):
    """Form field for a :py:class:`~cg:cryptography.x509.CertificatePolicies` extension."""

    extension_type = x509.CertificatePolicies
    fields = (
        ObjectIdentifierField(required=False),  # Policy Identifier
        forms.CharField(required=False),
        forms.CharField(required=False),
    )
    widget = widgets.CertificatePoliciesWidget

    def get_value(  # type: ignore[override]
        self, policy_identifier: x509.ObjectIdentifier, practice_statements: str, explicit_text: str
    ) -> Optional[x509.CertificatePolicies]:
        if not policy_identifier or not (practice_statements or explicit_text):
            return None

        policy_qualifiers = typing.cast(List[Union[str, x509.UserNotice]], practice_statements.splitlines())
        if explicit_text:
            policy_qualifiers.append(x509.UserNotice(notice_reference=None, explicit_text=explicit_text))

        policy_information = x509.PolicyInformation(
            policy_identifier=policy_identifier, policy_qualifiers=policy_qualifiers
        )
        return x509.CertificatePolicies([policy_information])


class CRLDistributionPointField(DistributionPointField[x509.CRLDistributionPoints]):
    """Form field for a :py:class:`~cg:cryptography.x509.CRLDistributionPoints` extension."""

    extension_type = x509.CRLDistributionPoints
    widget = widgets.CRLDistributionPointsWidget


class ExtendedKeyUsageField(MultipleChoiceExtensionField[x509.ExtendedKeyUsage]):
    """Form field for a :py:class:`~cg:cryptography.x509.ExtendedKeyUsage` extension."""

    extension_type = x509.ExtendedKeyUsage
    choices = _EXTENDED_KEY_USAGE_CHOICES
    widget = widgets.ExtendedKeyUsageWidget

    def get_values(self, value: List[str]) -> Optional[x509.ExtendedKeyUsage]:
        return x509.ExtendedKeyUsage(usages=[_EXTENDED_KEY_USAGE_MAPPING[name] for name in value])


class FreshestCRLField(DistributionPointField[x509.FreshestCRL]):
    """Form field for a :py:class:`~cg:cryptography.x509.CRLDistributionPoints` extension."""

    extension_type = x509.FreshestCRL
    widget = widgets.FreshestCRLWidget


class IssuerAlternativeNameField(ExtensionField[x509.IssuerAlternativeName]):
    """Form field for a :py:class:`~cg:cryptography.x509.IssuerAlternativeName` extension."""

    extension_type = x509.IssuerAlternativeName
    fields = (GeneralNamesField(required=False),)
    widget = widgets.IssuerAlternativeNameWidget

    def get_value(  # type: ignore[override]
        self, value: List[x509.GeneralName]
    ) -> Optional[x509.IssuerAlternativeName]:
        if not value:
            return None
        return x509.IssuerAlternativeName(general_names=value)


class KeyUsageField(MultipleChoiceExtensionField[x509.KeyUsage]):
    """Form field for a :py:class:`~cg:cryptography.x509.KeyUsage` extension."""

    choices = sorted(KEY_USAGE_NAMES.items(), key=lambda t: t[1])

    extension_type = x509.KeyUsage
    widget = widgets.KeyUsageWidget

    def get_values(self, value: List[str]) -> Optional[x509.KeyUsage]:
        values = {choice: choice in value for choice in KEY_USAGE_NAMES}
        return x509.KeyUsage(**values)


class OCSPNoCheckField(ExtensionField[x509.OCSPNoCheck]):
    """Form field for a :py:class:`~cg:cryptography.x509.OCSPNoCheck` extension."""

    extension_type = x509.OCSPNoCheck
    fields = (forms.BooleanField(required=False),)
    widget = widgets.OCSPNoCheckWidget

    def get_value(self, value: bool) -> Optional[x509.OCSPNoCheck]:  # type: ignore[override]
        if value is True:
            return self.extension_type()
        return None


class SubjectAlternativeNameField(ExtensionField[x509.SubjectAlternativeName]):
    """Form field for a :py:class:`~cg:cryptography.x509.SubjectAlternativeName` extension."""

    extension_type = x509.SubjectAlternativeName
    fields = (
        GeneralNamesField(required=False),
        forms.BooleanField(required=False),
    )
    widget = widgets.SubjectAlternativeNameWidget

    def compress(  # type: ignore[override]  # this is a special case
        self, data_list: List[Any]
    ) -> Tuple[Optional[x509.Extension[ExtensionTypeTypeVar]], bool]:
        default_cn_in_san = ca_settings.CA_PROFILES[ca_settings.CA_DEFAULT_PROFILE]["cn_in_san"]
        if not data_list:  # pragma: no cover
            return None, default_cn_in_san

        *value, critical = data_list
        ext_value, cn_in_san = self.get_value(*value)
        if ext_value is None:
            return None, cn_in_san
        ext = x509.Extension(critical=critical, oid=self.extension_type.oid, value=ext_value)
        # TYPE NOTE: mypy complains about the non-generic type being returned
        return ext, cn_in_san  # type: ignore[return-value]

    def get_value(  # type: ignore[override]
        self, names: List[x509.GeneralName], cn_in_san: bool
    ) -> Tuple[Optional[x509.SubjectAlternativeName], bool]:
        if not names:
            return None, cn_in_san
        return x509.SubjectAlternativeName(general_names=names), cn_in_san


class TLSFeatureField(MultipleChoiceExtensionField[x509.TLSFeature]):
    """Form field for a :py:class:`~cg:cryptography.x509.TLSFeature` extension."""

    extension_type = x509.TLSFeature
    choices = (
        (x509.TLSFeatureType.status_request.name, "status_request (OCSPMustStaple)"),
        (x509.TLSFeatureType.status_request_v2.name, "status_request_v2 (MultipleCertStatusRequest)"),
    )  # TODO: choices can also be a function - better for testing for completeness
    widget = widgets.TLSFeatureWidget

    def get_values(self, value: List[str]) -> Optional[x509.TLSFeature]:
        # Note: sort value to get predictable output in test cases
        features = [getattr(x509.TLSFeatureType, elem) for elem in sorted(value)]
        return self.extension_type(features=features)



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca. If not, see
# <http://www.gnu.org/licenses/>.

"""Management command to list all available certificates.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

from typing import Any, Optional

from django.core.management.base import CommandParser
from django.utils import timezone

from django_ca.management.base import BaseCommand
from django_ca.models import Certificate, CertificateAuthority
from django_ca.utils import add_colons


class Command(BaseCommand):  # pylint: disable=missing-class-docstring
    help = "List all certificates."

    def add_arguments(self, parser: CommandParser) -> None:
        self.add_ca(parser, no_default=True, help_text="Only output certificates by the named authority.")
        parser.add_argument(
            "--expired", default=False, action="store_true", help="Also list expired certificates."
        )
        parser.add_argument(
            "--autogenerated",
            default=False,
            action="store_true",
            help="Also list automatically generated certificates.",
        )
        parser.add_argument(
            "--revoked", default=False, action="store_true", help="Also list revoked certificates."
        )

    def handle(
        self,
        ca: Optional[CertificateAuthority],
        expired: bool,
        revoked: bool,
        autogenerated: bool,
        **options: Any,
    ) -> None:
        certs = Certificate.objects.order_by("expires", "cn", "serial")

        if expired is False:
            certs = certs.filter(expires__gt=timezone.now())
        if revoked is False:
            certs = certs.filter(revoked=False)
        if autogenerated is False:
            certs = certs.filter(autogenerated=False)

        if ca is not None:
            certs = certs.filter(ca=ca)

        for cert in certs:
            if cert.revoked is True:
                info = "revoked"
            else:
                word = "expires"
                if cert.expires < timezone.now():
                    word = "expired"

                info = f"{word}: {cert.expires.strftime('%Y-%m-%d')}"
            self.stdout.write(f"{add_colons(cert.serial)} - {cert.cn} ({info})")
